from ..api import op, use
from ..weave_types import Function, TimestampBinType
from .. import weave_types as types
from .. import graph
from .. import weave_internal
from .dict import dict_

from ..weave_internal import define_fn, call_fn, make_const_node

NICE_BIN_SIZES_SEC = [
    # TODO: will need more steps along here for smooth zooming.
    1e-9,  # ns
    1e-6,  # microsec
    1e-3,  # ms
    1,
    2.5,
    5,
    10,
    15,
    30,
    60,  # 1min
    300,  # 5min
    600,  # 10min
    1200,  # 20min
    1800,  # 30 min
    *(3600 * i for i in range(1, 25)),  # 1 - 24 hr, increments of 1hr
    *(86400 * i for i in range(2, 31)),  # 2 - 30 days, increments of 1 day
    *(
        86400 * 30 * i for i in range(2, 13)
    ),  # 2 - 12 months (assuming 1 month = 30days) increments of 1 month
    *(
        365 * 86400 * i for i in range(1, 11)
    ),  # 1 - 10 years (assuming 1 year = 365 days) increments of 1 year
]

NICE_BIN_SIZES_SEC_NODE = weave_internal.make_const_node(
    types.List(types.Number()), NICE_BIN_SIZES_SEC
)


@op(
    output_type=Function(
        input_types={"ts": types.Timestamp()},
        output_type=TimestampBinType,  # type: ignore
    ),
)
def timestamp_bins_fixed(bin_size_s: float):
    def body(ts):
        if bin_size_s <= 0:
            raise ValueError("bin_size_s must be greater than zero.")
        return dict_(start=ts.floor(bin_size_s), stop=ts.ceil(bin_size_s))

    # this should be vectorized by expansion
    return define_fn({"ts": types.Timestamp()}, body)


@op(
    name="timestamps-binsnice",
    input_type={
        "arr": types.List(types.optional(types.Timestamp())),
        "target_n_bins": types.Number(),
    },
    output_type=Function(
        input_types={"ts": types.Timestamp()}, output_type=TimestampBinType  # type: ignore
    ),
    render_info={"type": "function"},
)
def timestamp_bins_nice(arr, target_n_bins):
    arr_min = min(arr) if len(arr) > 0 else 0
    arr_max = max(arr) if len(arr) > 0 else 0
    exact_bin_size = ((arr_max - arr_min) / target_n_bins).total_seconds()  # type: ignore
    bin_size_s = min(NICE_BIN_SIZES_SEC, key=lambda x: abs(x / exact_bin_size - 1))
    return use(timestamp_bins_fixed(bin_size_s))


@op(
    name="timestamp-bin",
    input_type={
        "in_": types.Timestamp(),
        "bin_fn": types.optional(
            Function(
                input_types={"row": types.Timestamp()}, output_type=TimestampBinType  # type: ignore
            )
        ),
    },
    output_type=TimestampBinType,  # type: ignore
)
def timestamp_bin(in_, bin_fn):
    if not isinstance(bin_fn, graph.Node) and bin_fn == None:
        return None
    return use(call_fn(bin_fn.val, {"ts": make_const_node(types.Timestamp(), in_)}))



import datetime
import weave
from . import eval
from weave.timestamp import tz_aware_dt


def test_data():
    model = eval.EvalModel("a")

    proc = eval.PredictionProcess("000", model, [])
    pred1 = eval.Prediction(datetime.datetime(2022, 10, 8, 10, 26), 0, 3)
    proc.predictions.append(pred1)
    pred2 = eval.Prediction(datetime.datetime(2022, 10, 8, 10, 31), 1, 4)
    proc.predictions.append(pred2)
    weave.save(proc, name="proc-a:latest")

    proc = eval.PredictionProcess("000", model, [])
    pred3 = eval.Prediction(datetime.datetime(2022, 10, 8, 10, 39), 9, 10)
    proc.predictions.append(pred3)
    weave.save(proc, name="proc-b:latest")

    res = weave.use(
        model.all_predictions(
            datetime.datetime(2022, 10, 8, 10, 30),
            datetime.datetime(2022, 10, 8, 10, 40),
        )
    )

    # The weave system will always return tz-aware datetimes, so we need to
    # convert the expected values to be tz-aware as well.
    pred2.timestamp = tz_aware_dt(pred2.timestamp)
    pred3.timestamp = tz_aware_dt(pred3.timestamp)
    assert res == [pred2, pred3]



from __future__ import annotations
from typing_extensions import (
    NotRequired,
    TypedDict,
)

# In own file so we can use future annotations (NotRequired)


class PanelInfo(TypedDict):
    hidden: NotRequired[bool]



import os
import logging
import datetime
import json
import typing
from wandb import errors as wandb_errors
from wandb.apis.public import Run
from wandb.sdk.internal.file_pusher import FilePusher
from wandb.sdk.internal import file_stream
from wandb.sdk.internal.internal_api import Api as InternalApi
from wandb.sdk.lib import runid
from weave import wandb_client_api
from weave import errors

logger = logging.getLogger(__name__)


# We disable urllib warnings because they are noisy and not actionable.
logging.getLogger("urllib3").setLevel(logging.ERROR)
logging.getLogger("urllib3.connectionpool").setLevel(logging.ERROR)

WANDB_HIDDEN_JOB_TYPE = "sweep-controller"


class InMemoryLazyLiteRun:
    # ID
    _entity_name: str
    _project_name: str
    _run_name: str
    _step: int = 0

    # Optional
    _display_name: typing.Optional[str] = None
    _job_type: typing.Optional[str] = None
    _group: typing.Optional[str] = None

    # Property Cache
    _i_api: typing.Optional[InternalApi] = None
    _run: typing.Optional[Run] = None
    _stream: typing.Optional[file_stream.FileStreamApi] = None
    _pusher: typing.Optional[FilePusher] = None
    _use_async_file_stream: bool = False

    def __init__(
        self,
        entity_name: str,
        project_name: str,
        run_name: typing.Optional[str] = None,
        *,
        job_type: typing.Optional[str] = None,
        group: typing.Optional[str] = None,
        _hide_in_wb: bool = False,
        _use_async_file_stream: bool = False,
    ):
        wandb_client_api.assert_wandb_authenticated()

        # Technically, we could use the default entity and project here, but
        # heeding Shawn's advice, we should be explicit about what we're doing.
        # We can always move to the default later, but we can't go back.
        if entity_name is None or entity_name == "":
            raise ValueError(f"Must specify entity_name")
        elif project_name is None or project_name == "":
            raise ValueError(f"Must specify project_name")

        self._entity_name = entity_name
        self._project_name = project_name
        self._display_name = run_name
        self._run_name = run_name or runid.generate_id()
        self._job_type = job_type if not _hide_in_wb else WANDB_HIDDEN_JOB_TYPE
        if _hide_in_wb and group is None:
            group = "weave_hidden_runs"
        self._group = group

        self._use_async_file_stream = (
            _use_async_file_stream
            and os.getenv("WEAVE_DISABLE_ASYNC_FILE_STREAM") == None
        )

    def ensure_run(self) -> Run:
        return self.run

    @property
    def i_api(self) -> InternalApi:
        if self._i_api is None:
            self._i_api = InternalApi(
                {"project": self._project_name, "entity": self._entity_name}
            )
        return self._i_api

    @property
    def run(self) -> Run:
        if self._run is None:
            try:
                # Ensure project exists
                self.i_api.upsert_project(
                    project=self._project_name, entity=self._entity_name
                )

                # Produce a run
                run_res, _, _ = self.i_api.upsert_run(
                    name=self._run_name,
                    display_name=self._display_name,
                    job_type=self._job_type,
                    group=self._group,
                    project=self._project_name,
                    entity=self._entity_name,
                )

                self._run = Run(
                    wandb_client_api.wandb_public_api().client,
                    run_res["project"]["entity"]["name"],
                    run_res["project"]["name"],
                    run_res["name"],
                    {
                        "id": run_res["id"],
                        "config": "{}",
                        "systemMetrics": "{}",
                        "summaryMetrics": "{}",
                        "tags": [],
                        "description": None,
                        "notes": None,
                        "state": "running",
                    },
                )
            except wandb_errors.CommError as e:
                raise errors.WeaveWandbAuthenticationException()

            self.i_api.set_current_run_id(self._run.id)
            self._step = self._run.lastHistoryStep + 1

        return self._run

    @property
    def stream(self) -> file_stream.FileStreamApi:
        if self._stream is None:
            # Setup the FileStream
            self._stream = file_stream.FileStreamApi(
                self.i_api, self.run.id, datetime.datetime.utcnow().timestamp()
            )
            if self._use_async_file_stream:
                self._stream._client.headers.update(
                    {"X-WANDB-USE-ASYNC-FILESTREAM": "true"}
                )
            self._stream.set_file_policy(
                "wandb-history.jsonl",
                file_stream.JsonlFilePolicy(start_chunk_id=self._step),
            )
            self._stream.start()

        return self._stream

    @property
    def pusher(self) -> FilePusher:
        if self._pusher is None:
            self._pusher = FilePusher(self.i_api, self.stream)

        return self._pusher

    def log(self, row_dict: dict) -> None:
        stream = self.stream
        row_dict = {
            **row_dict,
            "_timestamp": datetime.datetime.utcnow().timestamp(),
        }
        if not self._use_async_file_stream:
            row_dict["_step"] = self._step
        self._step += 1
        stream.push("wandb-history.jsonl", json.dumps(row_dict))

    def finish(self) -> None:
        if self._stream is not None:
            # Finalize the run
            self.stream.finish(0)

        if self._pusher is not None:
            # Wait for the FilePusher and FileStream to finish
            self.pusher.finish()
            self.pusher.join()

        # Reset fields
        self._stream = None
        self._pusher = None
        self._run = None
        self._i_api = None
        self._step = 0

    def __del__(self) -> None:
        self.finish()



import dataclasses
import typing

import weave

from ... import panel
from ... import panel_util


@weave.op()
def single_distribution(input_node: weave.Node[list[float]]) -> weave.panels.Plot:
    binned = input_node.groupby(lambda v: round(v * 10) / 10).map(  # type: ignore
        lambda group: weave.ops.dict_(value=group.key(), count=group.count())
    )
    return weave.panels.Plot(
        binned,
        x=lambda row: row["value"],
        y=lambda row: row["count"],
        mark="bar",
    )


@weave.type()
class AdderConfig:
    operand: weave.Node[int] = dataclasses.field(
        default_factory=lambda: weave.graph.ConstNode(weave.types.Int(), 10)
    )


def adder_set_default_config(config, new_config):
    return new_config


# TODO: really annoying that I need the setter here.
@weave.op(setter=adder_set_default_config)
def adder_default_config(config: typing.Optional[AdderConfig]) -> AdderConfig:
    if config == None:
        return AdderConfig(operand=panel_util.make_node(0.1))
    return typing.cast(AdderConfig, config)


@weave.op()
def adder_config(
    input_node: weave.Node[int], config: AdderConfig
) -> weave.panels.LabeledItem:
    input_val = typing.cast(int, input_node)
    config = adder_default_config(config)
    return weave.panels.LabeledItem(
        label="operand",
        item=weave.panels.Slider(
            config=weave.panels.SliderConfig(weave.ops.execute(config.operand))
        ),
    )


@weave.op()
def adder(input_node: weave.Node[int], config: AdderConfig) -> weave.panels.LabeledItem:
    input_val = typing.cast(int, input_node)
    config = adder_default_config(config)
    return weave.panels.LabeledItem(label="output", item=input_val + config.operand)  # type: ignore


@weave.type()
class Adder(panel.Panel):
    id = "op-adder"
    config: typing.Optional[AdderConfig] = None

    def __init__(self, input_node, vars=None, config=None, **options):
        super().__init__(input_node=input_node, vars=vars)
        self.config = config
        if self.config is None:
            self.config = AdderConfig()

        if "operand" in options:
            self.config.operand = panel_util.make_node(options["operand"])



import typing
import pyarrow as pa
from pyarrow import compute as pc

from ..language_features.tagging import tag_store, process_opdef_output_type
from .. import weave_types as types

from .arrow import offsets_starting_at_zero
from . import convert

if typing.TYPE_CHECKING:
    from list_ import ArrowWeaveList  # type: ignore[import]


def recursively_encode_pyarrow_strings_as_dictionaries(array: pa.Array) -> pa.Array:
    if pa.types.is_struct(array.type):
        if array.type.num_fields == 0:
            return array
        return pa.StructArray.from_arrays(
            [
                recursively_encode_pyarrow_strings_as_dictionaries(
                    array.field(field.name)
                )
                for field in array.type
            ],
            [field.name for field in array.type],
            mask=pa.compute.invert(array.is_valid()),
        )
    elif pa.types.is_list(array.type):
        return pa.ListArray.from_arrays(
            offsets_starting_at_zero(array),
            recursively_encode_pyarrow_strings_as_dictionaries(array.flatten()),
            mask=pa.compute.invert(array.is_valid()),
        )
    elif array.type == pa.string():
        return pc.dictionary_encode(array)
    else:
        return array


def direct_add_arrow_tags(
    data: typing.Union[pa.Table, pa.Array], arrow_tags: pa.StructArray
):
    arrow_tags = recursively_encode_pyarrow_strings_as_dictionaries(arrow_tags)
    current_tags = None
    if isinstance(data, pa.Table):
        if "_tag" in data.column_names:
            current_tags = data["_tag"].combine_chunks()
    elif isinstance(data, pa.StructArray):
        if data.type.get_field_index("_tag") > -1:
            current_tags = data.field("_tag")
    if current_tags is None:
        tag_arrays = []
        tag_names = []
    else:
        tag_arrays = [current_tags.field(f.name) for f in current_tags.type]
        tag_names = [f.name for f in current_tags.type]

    for tag_field in arrow_tags.type:
        # Don't overwrite tags that already exist, we want to keep the innermost ones!
        if tag_field.name not in tag_names:
            tag_arrays.append(arrow_tags.field(tag_field.name))
            tag_names.append(tag_field.name)

    tag_array = pa.StructArray.from_arrays(
        tag_arrays,
        tag_names,
    )
    if isinstance(data, pa.Table):
        if current_tags is not None:
            new_value = data["_value"]
        else:
            new_value = pa.StructArray.from_arrays(
                # TODO: we shouldn't need to combine chunks, we can produce this in the
                # original chunked form for zero copy
                [c.combine_chunks() for c in data.columns],
                names=data.column_names,
            )
    elif isinstance(data, pa.StructArray):
        if current_tags is not None:
            new_value = data.field("_value")
        else:
            new_value = data
    else:
        # Else its an arrow array
        new_value = data
    return pa.StructArray.from_arrays([tag_array, new_value], ["_tag", "_value"])


def tag_arrow_array_elements_with_single_tag_dict(
    array: pa.Array, py_tags: dict
) -> pa.StructArray:
    tag_no_dictionary = convert.to_arrow([py_tags])._arrow_data
    tag_maybe_dictionary_encoded = recursively_encode_pyarrow_strings_as_dictionaries(
        tag_no_dictionary
    )
    tags = pa.repeat(tag_maybe_dictionary_encoded[0], len(array))
    return direct_add_arrow_tags(array, tags)


def awl_add_arrow_tags(
    l: "ArrowWeaveList", arrow_tags: pa.StructArray, tag_type: types.Type
):
    data = l._arrow_data
    new_value = direct_add_arrow_tags(data, arrow_tags)
    new_object_type = process_opdef_output_type.op_make_type_tagged_resolver(
        l.object_type, tag_type
    )
    from .list_ import ArrowWeaveList

    res: ArrowWeaveList = ArrowWeaveList(new_value, new_object_type, l._artifact)
    if tag_store.is_tagged(l):
        res = tag_store.add_tags(res, tag_store.get_tags(l))
    return res


def tag_awl_list_elements_with_single_tag_dict(
    awl: "ArrowWeaveList", py_tags: dict
) -> "ArrowWeaveList":
    tag_type = types.TypeRegistry.type_of(py_tags)
    tag_array = tag_arrow_array_elements_with_single_tag_dict(
        awl._arrow_data, py_tags
    ).field("_tag")
    return awl_add_arrow_tags(awl, tag_array, tag_type)


def pushdown_list_tags(arr: "ArrowWeaveList") -> "ArrowWeaveList":
    if tag_store.is_tagged(arr):
        tag = tag_store.get_tags(arr)
        arr = tag_awl_list_elements_with_single_tag_dict(arr, tag)
    return arr



import contextlib
import dataclasses
import typing

from . import weave_types as types

TRACE_EXT = "trace.json"


def wb_object_type_from_path(path: str) -> typing.Tuple[types.Type, str]:
    parts = path.split(".")
    ext = ""
    wbObjectType: types.Type = types.NoneType()
    if path.endswith(TRACE_EXT):
        ext = TRACE_EXT
    elif len(parts) != 1:
        ext = parts[-1]
    if len(parts) > 2 and ext == "json":
        pathext_wbobjecttype = types.type_name_to_type(parts[-2])
        if pathext_wbobjecttype != None:
            wbObjectType = pathext_wbobjecttype()
    return wbObjectType, ext


@dataclasses.dataclass(frozen=True)
# It'd be nice to just name this FileType, but Weave0 uses that to mean
# ArtifactFile
class FileBaseType(types.Type):
    extension: types.Type = types.String()
    # Match weave0 snakeCase name.
    wbObjectType: types.Type = types.Any()

    @classmethod
    def type_of_instance(cls, obj: "File") -> "FileBaseType":
        # Default implementation for Types that take no arguments.
        wbObjectType, ext = wb_object_type_from_path(obj.path)
        return cls(
            extension=types.Const(types.String(), ext), wbObjectType=wbObjectType
        )

    def _to_dict(self) -> dict:
        # NOTE: js_compat
        # In the js Weave code, file is a non-standard type that
        # puts a const string at extension as just a plain string.
        d = super()._to_dict()
        if isinstance(self.extension, types.Const):
            d["extension"] = self.extension.val
        else:
            d.pop("extension")
        if not isinstance(self.wbObjectType, (types.Any, types.NoneType)):
            d["wbObjectType"] = self.wbObjectType.to_dict()
        else:
            d.pop("wbObjectType")
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "FileBaseType":
        # NOTE: js_compat
        # In the js Weave code, file is a non-standard type that
        # puts a const string at extension as just a plain string.
        extension: types.Type = types.String()
        if "extension" in d:
            extension = types.Const(types.String(), d["extension"])
        wbObjectType: types.Type = types.NoneType()
        if "wbObjectType" in d:
            wbObjectType = types.TypeRegistry.type_from_dict(d["wbObjectType"])
        return cls(extension, wbObjectType)


class File:
    path: str

    def size(self) -> int:
        raise NotImplementedError

    @contextlib.contextmanager
    def open(self, mode: str = "r") -> typing.Generator[typing.IO, None, None]:
        raise NotImplementedError

    def digest(self) -> typing.Optional[str]:
        raise NotImplementedError


FileBaseType.instance_classes = File


class BaseDirType(types.ObjectType):
    def property_types(self) -> dict[str, types.Type]:
        return {
            "fullPath": types.String(),
            "size": types.Int(),
            "dirs": types.Dict(types.String(), SubDirType()),
            "files": types.Dict(types.String(), FileBaseType()),
        }


class Dir:
    def __init__(
        self,
        fullPath: str,
        size: int,
        dirs: typing.Mapping[str, "SubDir"],
        files: typing.Mapping[str, File],
    ):
        self.fullPath = fullPath
        self.size = size
        self.dirs = dirs
        self.files = files

    def path_info(self, path: str) -> typing.Union[File, "Dir", None]:
        raise NotImplementedError


BaseDirType.instance_classes = Dir


@dataclasses.dataclass(frozen=True)
class SubDirType(types.ObjectType):
    # TODO doesn't match frontend
    name = "subdir"

    # A type argument, can be subdir of LocalFile or of FilesystemArtifactFile
    file_type: types.Type = FileBaseType()

    def property_types(self) -> dict[str, types.Type]:
        return {
            "fullPath": types.String(),
            "size": types.Int(),
            "dirs": types.Dict(types.String(), types.Int()),
            "files": types.Dict(types.String(), self.file_type),
        }


class SubDir:
    def __init__(
        self,
        fullPath: str,
        size: int,
        dirs: dict[str, int],
        files: typing.Mapping[str, File],
    ):
        self.fullPath = fullPath
        self.size = size
        self.dirs = dirs
        self.files = files


SubDirType.instance_classes = SubDir


# attach to types, a lot of places expect it there
types.FileType = FileBaseType  # type: ignore



import pyarrow as pa


##### Classes and methods for working with arrow Types


class ArrowTypeWithFieldInfo:
    def __init__(self, type_, nullable, metadata):
        self.type = type_
        self.nullable = nullable
        self.metadata = metadata


def arrow_type_with_metadata(type_, metadata):
    if isinstance(type_, ArrowTypeWithFieldInfo):
        # blow away existing
        type_.metadata = metadata
    else:
        return ArrowTypeWithFieldInfo(type_, False, metadata)


def arrow_type_with_nullable(type_):
    if isinstance(type_, ArrowTypeWithFieldInfo):
        type_.nullable = True
    else:
        return ArrowTypeWithFieldInfo(type_, True, None)


def arrow_field(name, type_):
    if isinstance(type_, ArrowTypeWithFieldInfo):
        return pa.field(
            name, type_.type, nullable=type_.nullable, metadata=type_.metadata
        )
    return pa.field(name, type_)


def arrow_type(type_):
    if isinstance(type_, ArrowTypeWithFieldInfo):
        return type_.type
    return type_



import base64
import io
import requests
from PIL import Image
import weave


class ServiceError(Exception):
    def __init__(self, status_code):
        self.status_code = status_code


@weave.op(render_info={"type": "function"})
def generate(prompt: str) -> list[Image.Image]:
    if prompt is None:
        return None
    r = requests.post("https://bf.dallemini.ai/generate", json={"prompt": prompt})
    if r.status_code == 200:
        json = r.json()
        images = json["images"]
        images = [Image.open(io.BytesIO(base64.b64decode(img))) for img in images]
        return images
    else:
        raise ServiceError(r.status_code)



from __future__ import annotations
from typing_extensions import (
    NotRequired,
    TypedDict,
)

import typing
import weave

import pyarrow as pa

from ... import ops_arrow

import hdbscan


class HDBSCANOptions(TypedDict):
    min_samples: NotRequired[int]
    min_cluster_size: NotRequired[int]


@weave.op()
def hdbscan_cluster(
    data: list[list[float]], options: typing.Optional[HDBSCANOptions]
) -> ops_arrow.ArrowWeaveList[typing.Optional[str]]:
    if options == None:
        options = {}  # type: ignore
    np_arr = hdbscan.HDBSCAN(**options).fit_predict(data)

    arr = pa.array(np_arr)
    item_is_null = pa.compute.equal(arr, pa.scalar(-1))
    result = pa.compute.replace_with_mask(
        arr, item_is_null, pa.scalar(None, pa.int64())
    )
    str_result = result.cast(pa.string())

    return ops_arrow.ArrowWeaveList(str_result)



import contextlib
import dataclasses
import typing
import json
import datetime
import os

from . import artifact_base
from . import ref_base
from . import ref_util
from . import weave_types as types
from . import uris
from . import errors
from . import file_base
from . import object_context

from .language_features.tagging import tag_store


class FilesystemArtifactType(types.Type):
    def save_instance(
        self, obj: "FilesystemArtifact", artifact: "FilesystemArtifact", name: str
    ) -> "FilesystemArtifactRef":
        return FilesystemArtifactRef(obj, None)


class BranchPointType(typing.TypedDict):
    commit: str
    n_commits: int
    original_uri: str


@dataclasses.dataclass
class ArtifactMetadata:
    """
    Simple class to represent metadata for an artifact. Since
    Artifacts can be any where between non-saved, saved, and mutated,
    their metadata itself is split into two parts: mutable and readonly.

    Mutable metadata is metadata takes precedence over readonly metadata.
    """

    _mutable_metadata: typing.Dict[str, typing.Any]
    _readonly_metadata: typing.Dict[str, typing.Any]

    def __getitem__(self, key: str) -> typing.Any:
        return self._mutable_metadata.get(key, self._readonly_metadata.get(key))

    def __setitem__(self, key: str, value: typing.Any) -> None:
        self._mutable_metadata[key] = value

    def __delitem__(self, key: str) -> None:
        del self._mutable_metadata[key]

    def __contains__(self, key: str) -> bool:
        return key in self._mutable_metadata or key in self._readonly_metadata

    def as_dict(self) -> typing.Dict[str, typing.Any]:
        return {**self._readonly_metadata, **self._mutable_metadata}

    def update(self, other: typing.Dict[str, typing.Any]) -> None:
        self._mutable_metadata.update(other)


class FilesystemArtifact(artifact_base.Artifact):
    RefClass: typing.ClassVar[typing.Type["FilesystemArtifactRef"]]
    name: str

    def set(
        self, key: str, type_: types.Type, obj: typing.Any
    ) -> artifact_base.ArtifactRef:
        # We should do this, but it doesn't work yet!
        # existing_ref = ref_base.get_ref(obj)
        # if isinstance(existing_ref, artifact_base.ArtifactRef):
        #     return existing_ref
        ref_extra = type_.save_instance(obj, self, key)
        # If save_instance returned a Ref, return that directly.
        # TODO: refactor
        if isinstance(ref_extra, ref_base.Ref):
            return ref_extra
        with self.new_file(f"{key}.type.json") as f:
            json.dump(type_.to_dict(), f)
        return self.RefClass(self, path=key, type=type_, obj=obj, extra=ref_extra)

    def get(self, key: str, type_: types.Type) -> typing.Any:
        return self.ref_from_local_str(key, type_).get()

    @property
    def is_saved(self) -> bool:
        raise NotImplementedError

    @contextlib.contextmanager
    def open(
        self, path: str, binary: bool = False
    ) -> typing.Generator[typing.IO, None, None]:
        raise NotImplementedError

    @contextlib.contextmanager
    def new_file(
        self, path: str, binary: bool = False
    ) -> typing.Generator[typing.IO, None, None]:
        raise NotImplementedError

    def ref_from_local_str(self, s: str, type: "types.Type") -> "FilesystemArtifactRef":
        path, extra = ref_util.parse_local_ref_str(s)
        return self.RefClass(self, path=path, extra=extra, type=type)

    @property
    def created_at(self) -> datetime.datetime:
        raise NotImplementedError

    @property
    def version(self) -> str:
        raise NotImplementedError

    @property
    def branch(self) -> typing.Optional[str]:
        raise NotImplementedError

    @property
    def branch_point(self) -> typing.Optional[BranchPointType]:
        raise NotImplementedError

    @property
    def initial_uri_obj(self) -> uris.WeaveURI:
        # uri_obj must be a stable identifier (no aliases) for this artifact.
        # initial_uri_obj is the uri used to construct the artifact, which may include
        # aliases. artifact_wandb is the only case where we need to override this
        # to provide different behaviors for each, because it does not resolve aliases
        # upon construction. storage.to_weavejs is the only place that makes use
        # of initial_uri_obj. It doesn't want to incur the cost of resolving potentially
        # many uris (as in a run summary object), and the calling code doesn't care.
        return self.uri_obj

    @property
    def uri_obj(self) -> uris.WeaveURI:
        raise NotImplementedError

    @property
    def uri(self) -> str:
        return str(self.uri_obj)

    @property
    def metadata(self) -> ArtifactMetadata:
        raise NotImplementedError

    def delete(self) -> None:
        raise NotImplementedError

    def rename(self, new_name: str) -> None:
        raise NotImplementedError

    def direct_url(self, path: str) -> typing.Optional[str]:
        raise NotImplementedError

    def path(self, path: str) -> str:
        raise NotImplementedError

    def size(self, path: str) -> int:
        return os.path.getsize(self.path(path))

    def path_info(
        self, path: str
    ) -> typing.Optional[
        typing.Union["FilesystemArtifactFile", "FilesystemArtifactDir"]
    ]:
        res = self._path_info(path)
        if isinstance(res, FilesystemArtifactRef):
            if res.path is None:
                raise errors.WeaveInternalError(f"Cannot get path info for {path}")
            return res.artifact.path_info(res.path)
        return res

    def _path_info(
        self, path: str
    ) -> typing.Optional[
        typing.Union[
            "FilesystemArtifactFile", "FilesystemArtifactDir", "FilesystemArtifactRef"
        ]
    ]:
        raise NotImplementedError

    def previous_uri(self) -> typing.Optional[str]:
        raise NotImplementedError

    def undo(self) -> typing.Optional["FilesystemArtifact"]:
        raise NotImplementedError


FilesystemArtifactType.instance_classes = FilesystemArtifact


class FilesystemArtifactRef(artifact_base.ArtifactRef):
    FileType: typing.ClassVar[typing.Type[types.Type]]

    artifact: FilesystemArtifact
    path: typing.Optional[str]

    @classmethod
    def from_uri(cls, uri: uris.WeaveURI) -> "FilesystemArtifactRef":
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"<{self.__class__}({id(self)}) artifact={self.artifact} path={self.path} type={self.type} extra={self.extra}"

    @property
    def _outer_type(self) -> types.Type:
        if self.path is None:
            return types.TypeRegistry.type_of(self.artifact)

        # This block is required to fix some W&B tests that mock artifacts and don't
        # provide a .type.json file. They rely on the ref being constructed with a
        # type. When _get is called, it calls _outer_type to try to read the artifact,
        # which needs to early return here to avoid the type.json read.
        if self.extra is None and self._type is not None:
            return self._type

        try:
            with self.artifact.open(f"{self.path}.type.json") as f:
                type_json = json.load(f)
        except FileNotFoundError:
            return types.TypeRegistry.type_of(self.artifact.path_info(self.path))
        return types.TypeRegistry.type_from_dict(type_json)

    @property
    def type(self) -> types.Type:
        obj_ctx = object_context.get_object_context()
        if obj_ctx:
            t = obj_ctx.lookup_ref_type(str(self))
            if not isinstance(t, object_context.NoResult):
                return t

        if self._type is not None:
            return self._type

        if not self.is_saved:
            return types.NoneType()

        ot = self._outer_type
        if self.extra is not None:
            # This is some hacking to make extra refs work (a ref that's a pointer to
            # an item in a list)...
            # Not a general solution!
            # TODO: fix
            from . import types_numpy

            if isinstance(ot, types.List):
                self._type = ot.object_type
            elif isinstance(ot, types_numpy.NumpyArrayType):
                # The Numpy type implementation is not consistent with how list extra refs
                # work!
                # TODO: fix
                self._type = ot
            else:
                print(
                    "SELF TYPE",
                )
                raise errors.WeaveInternalError(
                    f"Cannot get type of {self} with extra {self.extra}"
                )
        else:
            self._type = ot
        return self._type

    # Artifact Name (unique per project, unique per local-artifact user namespace)
    @property
    def name(self) -> str:
        return self.artifact.name

    # The identifier for the artifact (typically the commit_hash, but can be branch)
    @property
    def version(self) -> str:
        return self.artifact.version

    # An optional branchname by which this artifact is identified.
    @property
    def branch(self) -> typing.Optional[str]:
        return self.artifact.branch

    @property
    def branch_point(self) -> typing.Optional[BranchPointType]:
        return self.artifact.branch_point

    @property
    def is_saved(self) -> bool:
        return self.artifact.is_saved

    # The original URI used to construct the artifact
    @property
    def initial_uri(self) -> str:
        uri = typing.cast("FilesystemArtifactURI", self.artifact.initial_uri_obj)
        uri = typing.cast("FilesystemArtifactURI", uri.parse(str(uri)))
        uri.path = self.path
        uri.extra = self.extra
        return str(uri)

    # A stable, fully qualified URI for this artifact (always uses commit_hash for version)
    @property
    def uri(self) -> str:
        uri = typing.cast("FilesystemArtifactURI", self.artifact.uri_obj)
        uri = typing.cast("FilesystemArtifactURI", uri.parse(str(uri)))
        uri.path = self.path
        uri.extra = self.extra
        return str(uri)

    # A URI suitable to branch from (will include branchname if appropriate)
    @property
    def branch_uri(self) -> str:
        uri = typing.cast("FilesystemArtifactURI", self.artifact.uri_obj)
        uri = typing.cast("FilesystemArtifactURI", uri.parse(str(uri)))
        uri.path = self.path
        uri.extra = self.extra
        if self.branch is not None:
            uri.version = self.branch
        return str(uri)

    def versions(self) -> list["FilesystemArtifactRef"]:
        raise NotImplementedError

    def _get(self) -> typing.Any:
        if self.path is None:
            return self.artifact
        # If we not in the execution path, we don't have a tagging context.
        # But we need a tagging context to load objects that were tagged.
        # This call will only create one if we don't already have one.
        # TODO: caching tagged values is already problematic in other ways
        # since they break idempotency. When we figure that out, we may
        # be able to remove this.
        with tag_store.isolated_tagging_context():
            return self._outer_type.load_instance(
                self.artifact, self.path, extra=self.extra
            )

    def local_ref_str(self) -> str:
        if self.path is None:
            raise errors.WeaveInternalError("Cannot get local ref str for artifact")
        s = self.path
        if self.extra is not None:
            s += "#%s" % "/".join(self.extra)
        return s


@dataclasses.dataclass
class FilesystemArtifactURI(uris.WeaveURI):
    name: str
    version: str
    path: typing.Optional[str] = None
    extra: typing.Optional[list[str]] = None

    def __init__(*args: list[typing.Any], **kwargs: dict[str, typing.Any]) -> None:
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class FilesystemArtifactFileType(file_base.FileBaseType):
    # Note the name! When Weave0 uses "file" it means FilesystemArtifactFile.
    name = "file"

    def save_instance(
        self,
        obj: "FilesystemArtifactFile",
        artifact: FilesystemArtifact,
        name: str,
    ) -> FilesystemArtifactRef:
        return FilesystemArtifactRef(obj.artifact, obj.path)

    def load_instance(
        self,
        artifact: FilesystemArtifact,
        name: str,
        extra: typing.Optional[list[str]] = None,
    ) -> "FilesystemArtifactFile":
        return FilesystemArtifactFile(artifact, name)


@dataclasses.dataclass
class FilesystemArtifactFile(file_base.File):
    artifact: "FilesystemArtifact"
    path: str

    # Providing a _ref property means that get_ref on this object will return the
    # ref.
    @property
    def _ref(self) -> FilesystemArtifactRef:
        return FilesystemArtifactRef(self.artifact, self.path, obj=self)

    @contextlib.contextmanager
    def open(self, mode: str = "r") -> typing.Generator[typing.IO, None, None]:
        if "r" in mode:
            binary = False
            if "b" in mode:
                binary = True
            with self.artifact.open(self.path, binary=binary) as f:
                yield f
        else:
            raise NotImplementedError

    def size(self) -> int:
        return self.artifact.size(self.path)

    def digest(self) -> typing.Optional[str]:
        from weave.artifact_wandb import WandbArtifact

        if isinstance(self.artifact, WandbArtifact):
            # we can get the digest from the manifest (much faster)
            return self.artifact.digest(self.path)
        else:
            # This matches how WandB calculates digests for files
            from wandb.sdk.lib import hashutil  # type: ignore[attr-defined]

            return hashutil.md5_file_b64(self.artifact.path(self.path))


FilesystemArtifactFileType.instance_classes = FilesystemArtifactFile


class FilesystemArtifactDirType(file_base.BaseDirType):
    name = "dir"
    artifact: "FilesystemArtifactType" = FilesystemArtifactType()

    def property_types(self) -> dict[str, types.Type]:
        return {
            "artifact": self.artifact,
            "fullPath": types.String(),
            "size": types.Int(),
            "dirs": types.Dict(
                types.String(),
                file_base.SubDirType(FilesystemArtifactFileType()),
            ),
            "files": types.Dict(types.String(), FilesystemArtifactFileType()),
        }


@dataclasses.dataclass
class FilesystemArtifactDir(file_base.Dir):
    artifact: "FilesystemArtifact"
    fullPath: str
    size: int
    # TODO: these should be mappings instead!
    dirs: dict[str, file_base.SubDir]
    files: dict[str, FilesystemArtifactFile]

    def path_info(
        self, path: str
    ) -> typing.Union[FilesystemArtifactFile, "FilesystemArtifactDir", None]:
        target_path = self.fullPath
        if target_path != "" and path != "":
            target_path += "/"
        target_path += path

        return self.artifact.path_info(target_path)


FilesystemArtifactDirType.instance_classes = FilesystemArtifactDir


def update_weave_meta(wb_type: types.Type, artifact: FilesystemArtifact) -> None:
    panel_type = types.type_name_to_type("Panel")
    artifact.metadata["_weave_meta"] = {
        "is_weave_obj": True,
        "type_name": wb_type.name,
        "is_panel": panel_type and panel_type().assign_type(wb_type),
    }



import typing

import weave
from .. import weave_internal


@weave.type()
class PyBoardSeedBoardConfig:
    pass


@weave.op(  # type: ignore
    name="py_board-seed_board",
)
def seed_board(
    input_node: weave.Node[typing.Any],
    config: typing.Optional[PyBoardSeedBoardConfig] = None,
) -> weave.panels.Group:
    control_items = [
        weave.panels.GroupPanel(
            input_node,
            id="data",
        ),
    ]

    panels = [
        weave.panels.BoardPanel(
            weave_internal.make_var_node(input_node.type, "data"),
            id="table",
            layout=weave.panels.BoardPanelLayout(x=0, y=0, w=24, h=6),
        ),
    ]
    return weave.panels.Board(vars=control_items, panels=panels)



import typing
import datetime

from . import engine_trace
from . import wandb_api
from . import environment
from . import errors

statsd = engine_trace.statsd()  # type: ignore


def get_user_cache_key() -> typing.Optional[str]:
    ctx = wandb_api.get_wandb_api_context()
    if ctx is None:
        if environment.is_public():
            raise errors.WeaveAccessDeniedError("No user set in public environment")
        return None
    return ctx.user_id


# TODO: stats, test

CacheKeyType = typing.TypeVar("CacheKeyType")
CacheValueType = typing.TypeVar("CacheValueType")


class LruTimeWindowCache(typing.Generic[CacheKeyType, CacheValueType]):
    """A cache that stores values for a fixed amount of time.

    Respects the user cache key, so that different users don't share the same cache.
    """

    class NotFound:
        pass

    NOT_FOUND = NotFound()

    def __init__(
        self,
        max_age: datetime.timedelta,
        now_fn: typing.Callable[[], datetime.datetime] = datetime.datetime.now,
    ) -> None:
        self.max_age = max_age
        self._now_fn = now_fn

        # Items are time ordered, with oldest at the front.
        self._cache: dict[
            typing.Tuple[typing.Optional[str], CacheKeyType],
            typing.Tuple[datetime.datetime, CacheValueType],
        ] = {}

    def _full_key(
        self, key: CacheKeyType
    ) -> typing.Tuple[typing.Optional[str], CacheKeyType]:
        return (get_user_cache_key(), key)

    def _prune(self, now: datetime.datetime) -> None:
        for key, val in list(self._cache.items()):
            if now - val[0] > self.max_age:
                del self._cache[key]
            else:
                break
        statsd.gauge("weave.cache.size", len(self._cache))

    def get(self, key: CacheKeyType) -> typing.Union[NotFound, CacheValueType]:
        full_key = self._full_key(key)
        val = self._cache.get(full_key)
        if val is None:
            statsd.increment("weave.cache.miss")
            return self.NOT_FOUND
        # Set the value again to move it to the end of the cache
        self.set(key, val[1])
        statsd.increment("weave.cache.hit")
        return val[1]

    def set(self, key: CacheKeyType, value: CacheValueType) -> None:
        full_key = self._full_key(key)
        now = self._now_fn()
        if full_key in self._cache:
            # Delete so we move to the end of the cache
            del self._cache[full_key]
        self._cache[full_key] = (now, value)
        self._prune(now)



import typing
import pyarrow as pa
from pyarrow import compute as pc

from .. import weave_types as types
from ..decorator_arrow_op import arrow_op
from ..language_features.tagging import tagged_value_type

from .arrow import ArrowWeaveListType, arrow_as_array
from .list_ import ArrowWeaveList


def _arrowweavelistlist_listindex_output_type(input_types):
    self = input_types["self"]
    new_type = types.optional(self.object_type.object_type)
    if new_type == types.UnknownType():
        # if our object type was Unknown, it means it was an empty list.
        # indexing into that will always produce None
        new_type = types.NoneType()
    return new_type


@arrow_op(
    name="ArrowWeaveListList-listindex",
    input_type={
        "self": ArrowWeaveListType(
            types.optional(types.UnionType(types.List(), ArrowWeaveListType()))
        ),
        "index": types.optional(
            types.UnionType(
                types.Int(),
                types.List(types.optional(types.Int())),
                ArrowWeaveListType(types.optional(types.Int())),
            )
        ),
    },
    output_type=lambda input_types: ArrowWeaveListType(
        _arrowweavelistlist_listindex_output_type(input_types)
    ),
)
def listindex(self, index):
    a = arrow_as_array(self._arrow_data)
    # Not handling negative indexes at the moment
    if index == None:
        return ArrowWeaveList(
            pa.nulls(len(a)),
            types.NoneType(),
            self._artifact,
        )
    if isinstance(index, int):
        assert index >= 0
    else:
        index = arrow_as_array(index._arrow_data)
        assert pa.compute.all(pa.compute.greater_equal(index, 0))

    start_indexes = a.offsets[:-1]
    end_indexes = a.offsets[1:]
    take_indexes = pa.compute.add(start_indexes, index)
    oob = pa.compute.greater_equal(take_indexes, end_indexes)
    take_indexes = pa.compute.if_else(oob, None, take_indexes)
    result = a.flatten().take(take_indexes)
    return ArrowWeaveList(
        result,
        _arrowweavelistlist_listindex_output_type({"self": self}),
        self._artifact,
    )


def _list_op_output_object_type(input_types):
    self_type = input_types["self"]
    object_type = typing.cast(types.List, self_type.object_type)
    inner_object_type = object_type.object_type
    # TODO: union
    if isinstance(inner_object_type, tagged_value_type.TaggedValueType):
        return inner_object_type.value
    return inner_object_type


def list_dim_downresolver(
    self: ArrowWeaveList, arrow_operation_name: str, output_object_type=None
):
    a = arrow_as_array(self._arrow_data)
    values = a.flatten()

    object_type = typing.cast(types.List, self.object_type)
    if isinstance(object_type.object_type, tagged_value_type.TaggedValueType):
        values = values.field("_value")

    start_indexes = a.offsets[:-1]
    end_indexes = a.offsets[1:]
    non_0len = pa.compute.not_equal(start_indexes, end_indexes)
    t = pa.Table.from_arrays([a.value_parent_indices(), values], ["keys", "values"])
    g = t.group_by("keys")
    non_0len_agged = g.aggregate([("values", arrow_operation_name)])[
        f"values_{arrow_operation_name}"
    ]
    nulls = pa.nulls(len(a), type=non_0len_agged.type)
    result = pa.compute.replace_with_mask(
        nulls, non_0len, non_0len_agged.combine_chunks()
    )
    if output_object_type == None:
        output_object_type = _list_op_output_object_type(
            {"self": ArrowWeaveListType(self.object_type)}
        )
    return ArrowWeaveList(
        result,
        output_object_type,
        self._artifact,
    )


@arrow_op(
    name="ArrowWeaveListListNumber-listnumbercount",
    input_type={
        "self": ArrowWeaveListType(
            types.optional(
                types.UnionType(
                    types.List(),
                    ArrowWeaveListType(),
                )
            )
        ),
    },
    output_type=ArrowWeaveListType(types.Int()),
)
def list_numbers_count(self):
    return list_dim_downresolver(self, "count", types.Int())


@arrow_op(
    name="ArrowWeaveListListNumber-listnumbermax",
    input_type={
        "self": ArrowWeaveListType(
            types.optional(
                types.UnionType(
                    types.List(types.optional(types.Number())),
                    ArrowWeaveListType(types.optional(types.Number())),
                )
            )
        ),
    },
    output_type=lambda input_types: ArrowWeaveListType(
        _list_op_output_object_type(input_types)
    ),
)
def list_numbers_max(self):
    return list_dim_downresolver(self, "max")


@arrow_op(
    name="ArrowWeaveListListNumber-listnumbermin",
    input_type={
        "self": ArrowWeaveListType(
            types.optional(
                types.UnionType(
                    types.List(types.optional(types.Number())),
                    ArrowWeaveListType(types.optional(types.Number())),
                )
            )
        ),
    },
    output_type=lambda input_types: ArrowWeaveListType(
        _list_op_output_object_type(input_types)
    ),
)
def list_numbers_min(self):
    return list_dim_downresolver(self, "min")


@arrow_op(
    name="ArrowWeaveListListNumber-listnumbersum",
    input_type={
        "self": ArrowWeaveListType(
            types.optional(
                types.UnionType(
                    types.List(types.optional(types.Number())),
                    ArrowWeaveListType(types.optional(types.Number())),
                )
            )
        ),
    },
    output_type=lambda input_types: ArrowWeaveListType(
        _list_op_output_object_type(input_types)
    ),
)
def list_numbers_sum(self):
    return list_dim_downresolver(self, "sum")


@arrow_op(
    name="ArrowWeaveListListNumber-listnumberavg",
    input_type={
        "self": ArrowWeaveListType(
            types.optional(
                types.UnionType(
                    types.List(types.optional(types.Number())),
                    ArrowWeaveListType(types.optional(types.Number())),
                )
            )
        ),
    },
    output_type=lambda input_types: ArrowWeaveListType(
        _list_op_output_object_type(input_types)
    ),
)
def list_numbers_avg(self):
    return list_dim_downresolver(self, "mean")


# This iterates over the outer list in python but keeps the inner comparisons vectorized
# in arrow.
def compare_sublists(array1: pa.ListArray, array2: pa.ListArray) -> pa.Array:
    if len(array1) != len(array2):
        raise ValueError("The input arrays have different lengths")

    result = []
    flattened1 = array1.flatten()
    flattened2 = array2.flatten()
    offsets1 = array1.offsets.to_pylist()
    offsets2 = array2.offsets.to_pylist()

    for i in range(len(array1)):
        start1, end1 = offsets1[i], offsets1[i + 1]
        start2, end2 = offsets2[i], offsets2[i + 1]
        sublist1 = flattened1.slice(start1, end1 - start1)
        sublist2 = flattened2.slice(start2, end2 - start2)
        result.append(sublist1.equals(sublist2))

    return result


@arrow_op(
    name="ArrowWeaveListListNumber-equal",
    input_type={
        "self": ArrowWeaveListType(
            types.optional(
                types.UnionType(
                    types.List(types.optional(types.Number())),
                    ArrowWeaveListType(types.optional(types.Number())),
                )
            )
        ),
        "other": ArrowWeaveListType(
            types.optional(
                types.UnionType(
                    types.List(types.optional(types.Number())),
                    ArrowWeaveListType(types.optional(types.Number())),
                )
            )
        ),
    },
    output_type=lambda input_types: ArrowWeaveListType(types.Boolean()),
)
def list_equal(self, other):
    return ArrowWeaveList(
        pa.array(
            compare_sublists(
                self.without_tags()._arrow_data, other.without_tags()._arrow_data
            )
        ),
        types.Boolean(),
        self._artifact,
    )


@arrow_op(
    name="ArrowWeaveListListNumber-notEqual",
    input_type={
        "self": ArrowWeaveListType(
            types.optional(
                types.UnionType(
                    types.List(types.optional(types.Number())),
                    ArrowWeaveListType(types.optional(types.Number())),
                )
            )
        ),
        "other": ArrowWeaveListType(
            types.optional(
                types.UnionType(
                    types.List(types.optional(types.Number())),
                    ArrowWeaveListType(types.optional(types.Number())),
                )
            )
        ),
    },
    output_type=lambda input_types: ArrowWeaveListType(types.Boolean()),
)
def list_not_equal(self, other):
    return ArrowWeaveList(
        pc.invert(list_equal.resolve_fn(self, other)._arrow_data),
        types.Boolean(),
        self._artifact,
    )


def _vectorized_dropna_object_type(
    outer_object_type: typing.Union[types.List, ArrowWeaveListType, types.UnionType]
) -> types.Type:
    outer_is_optional = types.is_optional(outer_object_type)
    if outer_is_optional:
        container_type = types.non_none(outer_object_type)
    else:
        container_type = outer_object_type
    container_class = typing.cast(
        typing.Union[typing.Type[ArrowWeaveListType], typing.Type[types.List]],
        container_type.__class__,
    )
    element_type = typing.cast(
        typing.Union[types.List, ArrowWeaveListType], container_type
    ).object_type

    ret_type = container_class(types.non_none(element_type))
    if outer_is_optional:
        ret_type = types.optional(ret_type)  # type: ignore
    return ret_type


@arrow_op(
    name="ArrowWeaveListList-vectorizedDropna",
    input_type={
        "self": ArrowWeaveListType(types.List()),
    },
    output_type=lambda input_types: ArrowWeaveListType(
        _vectorized_dropna_object_type(input_types["self"].object_type)
    ),
)
def dropna(self):
    a = arrow_as_array(self._arrow_data)

    # NOTE: does not work for unions right now because of pyarrow bugs for null union values
    # e.g., pyarrow.compute.is_null(UnionScalar<None>) segfaults
    # pyarrow.compute.is_null(UnionArray<1, 2, None, None, "a">) returns [True, True, True, True, True]
    start_indexes = a.offsets[:-1]
    end_indexes = a.offsets[1:]
    flattened = a.flatten()
    non_null = pa.compute.cast(
        pa.compute.invert(pa.compute.is_null(flattened)), a.offsets.type
    )
    new_data = pa.compute.drop_null(flattened)
    cumulative_non_null_counts = pa.compute.cumulative_sum(non_null)
    new_offsets = cumulative_non_null_counts.take(pa.compute.subtract(end_indexes, 1))
    new_offsets = pa.concat_arrays([start_indexes[:1], new_offsets])
    unflattened = pa.ListArray.from_arrays(
        new_offsets, new_data, mask=pa.compute.is_null(a)
    )

    return ArrowWeaveList(
        unflattened,
        _vectorized_dropna_object_type(self.object_type),
        self._artifact,
    )



import dataclasses

from .. import weave_types as types
from .. import api as weave


class HtmlType(types.Type):
    def save_instance(self, obj, artifact, name):
        with artifact.new_file(f"{name}.html") as f:
            f.write(obj.html)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.html") as f:
            return Html(f.read())


@weave.weave_class(weave_type=HtmlType)
@dataclasses.dataclass
class Html:
    html: str


HtmlType.instance_classes = Html



import os
import re
import typing
import pathlib
import functools

from . import errors
from . import ref_base
from . import artifact_base
from . import artifact_mem
from . import artifact_fs
from . import artifact_local
from . import artifact_wandb
from . import weave_types as types
from . import mappers_python
from . import box
from . import errors
from . import graph

Ref = ref_base.Ref

if typing.TYPE_CHECKING:
    from weave.wandb_interface.wandb_lite_run import InMemoryLazyLiteRun


def split_path_dotfile(path, dotfile_name):
    while path != "/":
        path, tail = os.path.split(path)
        if os.path.exists(os.path.join(path, dotfile_name)):
            return path, tail
    raise FileNotFoundError


def _get_name(wb_type: types.Type, obj: typing.Any) -> str:
    return wb_type.name
    # This tries to figure out which variable references obj.
    # But it is slow when there are a lot of references. If we want to do
    # something like this, we'll need to do it somewhere closer to user
    # interaction.
    # obj_names = util.find_names(obj)
    # return f"{wb_type.name}-{obj_names[-1]}"


def _get_weave_type(obj: typing.Any):
    try:
        return types.TypeRegistry.type_of(obj)
    except errors.WeaveTypeError as e:
        raise errors.WeaveSerializeError(
            "weave type error during serialization for object: %s. %s"
            % (obj, str(e.args))
        )


def _ensure_object_components_are_published(
    obj: typing.Any, wb_type: types.Type, artifact: artifact_wandb.WandbArtifact
):
    from weave.mappers_publisher import map_to_python_remote

    mapper = map_to_python_remote(wb_type, artifact)
    return mapper.apply(obj)


def _assert_valid_name_part(part: typing.Optional[str] = None):
    if part is None:
        return
    if not re.match(r"^[a-zA-Z0-9_\-]+$", part):
        raise ValueError(
            "Invalid name part %s. Must be alphanumeric, dashes, or underscores." % part
        )


def _assert_valid_entity_name(part: typing.Optional[str] = None):
    if part is None:
        return
    if not re.match(r"^[a-z0-9_\-]+$", part):
        raise ValueError(
            "Invalid entity name %s. Must be lowercase, digits, dashes, or underscores."
            % part
        )


def _assert_valid_project_name(part: typing.Optional[str] = None):
    if part is None:
        return
    if len(part) > 128:
        raise ValueError("Invalid project name %s. Must be <= 128 characters." % part)
    if re.match(r"[\\#?%:]", part):
        raise ValueError(
            "Invalid project name %s. Can not contain \\, #, ?, %%, or :" % part
        )


def _assert_valid_artifact_name(part: typing.Optional[str] = None):
    if part is None:
        return
    if len(part) > 128:
        raise ValueError("Invalid artifact name %s. Must be <= 128 characters." % part)
    if not re.match(r"^[a-zA-Z0-9_\-.]+$", part):  # from W&B Artifacts
        raise ValueError(
            "Invalid artifact name %s. Must be alphanumeric, periods, dashes, or underscores."
            % part
        )


def _direct_publish(
    obj: typing.Any,
    name: typing.Optional[str] = None,
    wb_project_name: typing.Optional[str] = None,
    wb_artifact_type_name: typing.Optional[str] = None,
    wb_entity_name: typing.Optional[str] = None,
    branch_name: typing.Optional[str] = None,
    metadata: typing.Optional[typing.Dict[str, typing.Any]] = None,
    assume_weave_type: typing.Optional[types.Type] = None,
    *,
    _lite_run: typing.Optional["InMemoryLazyLiteRun"] = None,
):
    weave_type = assume_weave_type or _get_weave_type(obj)

    wb_project_name = wb_project_name or artifact_wandb.DEFAULT_WEAVE_OBJ_PROJECT
    name = name or _get_name(weave_type, obj)
    wb_artifact_type_name = wb_artifact_type_name or weave_type.name

    _assert_valid_artifact_name(name)
    _assert_valid_project_name(wb_project_name)
    _assert_valid_name_part(wb_artifact_type_name)
    _assert_valid_name_part(branch_name)
    _assert_valid_entity_name(wb_entity_name)
    # Validate entity name once we have them

    obj = box.box(obj)
    artifact = artifact_wandb.WandbArtifact(name, type=wb_artifact_type_name)
    artifact.metadata.update(metadata or {})
    obj = _ensure_object_components_are_published(obj, weave_type, artifact)
    artifact_fs.update_weave_meta(weave_type, artifact)
    ref = artifact.set("obj", weave_type, obj)

    # Only save if we have a ref into the artifact we created above. Otherwise
    #     nothing new was created, so just return the existing ref.
    if ref.artifact == artifact:
        artifact.save(
            project=wb_project_name,
            entity_name=wb_entity_name,
            branch=branch_name,
            _lite_run=_lite_run,
        )

    return ref


def _direct_save(
    obj: typing.Any,
    name: typing.Optional[str] = None,
    branch_name: typing.Optional[str] = None,
    source_branch_name: typing.Optional[str] = None,
    assume_weave_type: typing.Optional[types.Type] = None,
    artifact: typing.Optional[artifact_local.LocalArtifact] = None,
):
    weave_type = assume_weave_type or _get_weave_type(obj)
    name = name or _get_name(weave_type, obj)

    _assert_valid_name_part(name)
    _assert_valid_name_part(branch_name)
    _assert_valid_name_part(source_branch_name)

    obj = box.box(obj)
    if artifact is None:
        # Using `version=source_branch_name` feels too magical. Would be better
        # to have a classmethod on LocalArtifact that takes a branch name and
        # returns an artifact with that branch name. This also precludes
        # directly saving an artifact with a branchpoint that is not local
        artifact = artifact_local.LocalArtifact(name, version=source_branch_name)
        artifact_fs.update_weave_meta(weave_type, artifact)
    ref = artifact.set("obj", weave_type, obj)

    # Only save if we have a ref into the artifact we created above. Otherwise
    #     nothing new was created, so just return the existing ref.
    if ref.artifact == artifact:
        artifact.save(branch=branch_name)

    return ref


def publish(obj, name=None, type=None):
    # We would probably refactor this method to be more like _direct_publish. This effectively
    # just a wrapper that let's the user specify project name with a slash.
    # TODO: should we only expose save for our API with a "remote" flag or something
    project_name = None
    if name is not None and "/" in name:
        project_name, name = name.split("/")

    return _direct_publish(
        obj,
        name=name,
        wb_project_name=project_name,
        assume_weave_type=type,
    )


def save(
    obj,
    name=None,
    type=None,
    artifact=None,
    branch=None,
) -> artifact_local.LocalArtifactRef:
    # We would probably refactor this method to be more like _direct_save. This effectively
    # just a wrapper that let's the user specify source_branch name with a slash.
    source_branch = None
    if name is not None and ":" in name:
        name, source_branch = name.split(":", 1)

    return _direct_save(
        obj=obj,
        name=name,
        branch_name=branch,
        source_branch_name=source_branch,
        assume_weave_type=type,
        artifact=artifact,
    )


def get(uri_s: typing.Union[str, ref_base.Ref]) -> typing.Any:
    if isinstance(uri_s, ref_base.Ref):
        return uri_s.get()
    return ref_base.Ref.from_str(uri_s).get()


get_local_version_ref = artifact_local.get_local_version_ref
get_local_version = artifact_local.get_local_version


def deref(ref):
    if isinstance(ref, ref_base.Ref):
        return ref.get()
    return ref


def _get_ref(obj: typing.Any) -> typing.Optional[ref_base.Ref]:
    if isinstance(obj, ref_base.Ref):
        return obj
    return ref_base.get_ref(obj)


def clear_ref(obj):
    ref_base.clear_ref(obj)


get_ref = _get_ref


# Return all local artifacts.
# Warning: may iterate a lot of the filesystem!
def local_artifacts() -> typing.List[artifact_local.LocalArtifact]:
    result = []
    obj_paths = sorted(
        pathlib.Path(artifact_local.local_artifact_dir()).iterdir(),
        key=os.path.getctime,
    )
    for art_path in obj_paths:
        if os.path.basename(art_path) != "tmp":
            result.append(artifact_local.LocalArtifact(art_path.name, None))
    return result


def all_objects():
    result = []
    obj_paths = sorted(
        pathlib.Path(artifact_local.local_artifact_dir()).iterdir(),
        key=os.path.getctime,
    )
    for art_path in obj_paths:
        ref = artifact_local.get_local_version_ref(art_path.name, "latest")
        if ref is not None:
            result.append((ref.created_at, ref))
    return [r[1] for r in sorted(result)]


def objects(
    of_type: types.Type, alias: str = "latest"
) -> typing.List[artifact_local.LocalArtifactRef]:
    result = []
    for art_name in os.listdir(artifact_local.local_artifact_dir()):
        try:
            ref = artifact_local.get_local_version_ref(art_name, alias)
            if ref is not None:
                if of_type.assign_type(ref.type):
                    # TODO: Why did I have this here?
                    # obj = ref.get()
                    # if isinstance(ref.type, types.RunType) and obj.op_name == "op-objects":
                    #     continue
                    result.append((ref.artifact.created_at, ref))
        except errors.WeaveSerializeError:
            # This happens because we may not have loaded ecosystem stuff that we need
            # to deserialize
            continue
    # Sorted by created_at
    return [r[1] for r in sorted(result)]


def recursively_unwrap_arrow(obj):
    if getattr(obj, "to_pylist_notags", None):
        return obj.to_pylist_notags()
    if getattr(obj, "as_py", None):
        return obj.as_py()
    if isinstance(obj, graph.Node):
        return obj
    if isinstance(obj, dict):
        return {k: recursively_unwrap_arrow(v) for (k, v) in obj.items()}
    elif isinstance(obj, list):
        return [recursively_unwrap_arrow(item) for item in obj]
    return obj


def _default_ref_persister_artifact(
    type: types.Type, refs: typing.Iterable[artifact_base.ArtifactRef]
) -> artifact_base.Artifact:
    fs_art = artifact_local.LocalArtifact(type.name, "latest")
    # Save all the reffed objects into the new artifact.
    for mem_ref in refs:
        if mem_ref.path is not None and mem_ref._type is not None:
            fs_art.set(mem_ref.path, mem_ref._type, mem_ref._obj)
    fs_art.save()
    return fs_art


def to_python(
    obj: typing.Any,
    wb_type: typing.Optional[types.Type] = None,
    ref_persister: typing.Callable[
        [types.Type, typing.Iterable[artifact_base.ArtifactRef]], artifact_base.Artifact
    ] = _default_ref_persister_artifact,
) -> dict:
    if wb_type is None:
        wb_type = types.TypeRegistry.type_of(obj)

    # First map the object using a MemArtifact to capture any custom object refs.
    art = artifact_mem.MemArtifact()
    mapper = mappers_python.map_to_python(wb_type, art)
    val = mapper.apply(obj)

    if art.ref_count() > 0:
        # There are custom objects, create a local artifact to persist them.
        fs_art = ref_persister(wb_type, art.refs())
        # now map the original object again. Because there are now existing refs
        # to the local artifact for any custom objects, this new value will contain
        # those existing refs as absolute refs. We provide None for artifact because
        # it should not be used in this pass.
        # TODO: actually had to add artifact here to get HF/Bertviz caching
        # to work. Why?
        mapper = mappers_python.map_to_python(wb_type, fs_art)  # type: ignore
        val = mapper.apply(obj)
    # TODO: this should be a ConstNode!
    return {"_type": wb_type.to_dict(), "_val": val}


def to_safe_const(obj):
    wb_type = types.TypeRegistry.type_of(obj)
    mapper = mappers_python.map_to_python(wb_type, artifact_mem.MemArtifact())
    val = mapper.apply(obj)
    return graph.ConstNode(wb_type, val)


def from_python(obj: dict, wb_type: typing.Optional[types.Type] = None) -> typing.Any:
    if wb_type is None:
        wb_type = types.TypeRegistry.type_from_dict(obj["_type"])
    mapper = mappers_python.map_from_python(wb_type, None)  # type: ignore
    res = mapper.apply(obj["_val"])
    return res


def make_js_serializer():
    artifact = artifact_mem.MemArtifact()
    return functools.partial(to_weavejs, artifact=artifact)


def to_weavejs(obj, artifact: typing.Optional[artifact_base.Artifact] = None):
    from .ops_arrow import list_ as arrow_list

    obj = box.unbox(obj)
    if isinstance(obj, (str, int, float, bool, type(None))):
        return obj
    elif isinstance(obj, dict):
        return {k: to_weavejs(v, artifact=artifact) for (k, v) in obj.items()}
    elif isinstance(obj, list):
        return [to_weavejs(item, artifact=artifact) for item in obj]
    elif isinstance(obj, arrow_list.ArrowWeaveList):
        return obj.to_pylist_notags()
    wb_type = types.TypeRegistry.type_of(obj)

    if artifact is None:
        artifact = artifact_mem.MemArtifact()

    mapper = mappers_python.map_to_python(
        wb_type, artifact, mapper_options={"use_stable_refs": False}
    )
    return mapper.apply(obj)



import typing

from weave.op_args import OpNamedArgs

from . import op_def
from . import op_def_type
from . import op_args
from . import weave_types
from . import errors
from . import context_state
from . import storage
from . import uris
from . import op_aliases


class Registry:
    _types: typing.Dict[str, weave_types.Type]

    # This most recent register_op() call for a given OpDef.name
    # TODO: Get rid of this! Always use versioning! This is a temporary
    # state.
    _ops: typing.Dict[str, op_def.OpDef]

    # common_name: name: op_def
    _ops_by_common_name: typing.Dict[str, dict[str, op_def.OpDef]]

    _op_versions: typing.Dict[typing.Tuple[str, str], op_def.OpDef]

    def __init__(self):
        self._types = {}
        self._ops = {}
        self._ops_by_common_name = {}
        self._op_versions = {}

    def register_op(self, op: op_def.OpDef):
        # Always save OpDefs any time they are declared
        location = context_state.get_loading_op_location()
        is_loading = location is not None
        # do not save built-in ops today
        should_save = not is_loading and not op.is_builtin
        if should_save:
            # if we're not loading an existing op, save it.
            ref = storage.save(op, name=op.name + ":latest")
            version = ref.version
            location = ref.artifact.path_uri("obj")
        version = location.version if location is not None else None
        op.version = version
        op.location = location

        # if not is_loading:
        self._ops[op.name] = op
        self._ops_by_common_name.setdefault(op.common_name, {})[op.name] = op
        if version:
            self._op_versions[(op.name, version)] = op
        return op

    def have_op(self, op_name: str) -> bool:
        return op_name in self._ops

    def get_op(self, uri: str) -> op_def.OpDef:
        object_uri = uris.WeaveURI.parse(uri)
        if object_uri.version is not None:
            object_key = (object_uri.name, object_uri.version)
            if object_key in self._op_versions:
                res = self._op_versions[object_key]
            else:
                res = storage.get(uri)
                self._op_versions[object_key] = res
        elif object_uri.name in self._ops:
            res = self._ops[object_uri.name]
        else:
            if not ":" in uri:
                raise errors.WeaveInternalError("Op not registered: %s" % uri)
            res = storage.get(uri)
        if res is None:
            raise errors.WeaveInternalError("Op not registered: %s" % uri)
        return res

    def find_op_by_fn(self, lazy_local_fn):
        for op_def in self._op_versions.values():
            if op_def.call_fn == lazy_local_fn:
                return op_def
        raise Exception("Op def doesn't exist for %s" % lazy_local_fn)

    def find_ops_by_common_name(self, common_name: str) -> typing.List[op_def.OpDef]:
        aliases = op_aliases.get_op_aliases(common_name)
        ops: list[op_def.OpDef] = []
        for alias in aliases:
            ops.extend(self._ops_by_common_name.get(alias, {}).values())
        return ops

    def find_chainable_ops(
        self, arg0_type: weave_types.Type
    ) -> typing.List[op_def.OpDef]:
        def is_chainable(op):
            if not isinstance(op.input_type, op_args.OpNamedArgs):
                return False
            args = list(op.input_type.arg_types.values())
            if not args:
                return False
            return args[0].assign_type(arg0_type)

        return [op for op in self._ops.values() if is_chainable(op)]

    def load_saved_ops(self):
        for op_ref in storage.objects(op_def_type.OpDefType()):
            try:
                op_ref.get()
            except:
                # print("Failed to load non-builtin op: %s" % op_ref.uri)
                pass

    def list_ops(self) -> typing.List[op_def.OpDef]:
        # Note this uses self._ops, so provides the most recent registered op, which could
        # be the last one we loaded() [rather than the last one the user declared] which
        # is incorrect behavior
        return list(self._ops.values())

    # Currently this just returns all ops that take no arguments.
    # Perhaps a better extension is to require a return type that
    # subclasses some abstract package type?
    def list_packages(self) -> typing.List[op_def.OpDef]:
        packages = [
            a
            for a in list(self._ops.values())
            if isinstance(a.input_type, OpNamedArgs)
            and len(a.input_type.arg_types.keys()) == 0
        ]
        return packages

    def rename_op(self, name, new_name):
        """Internal use only, used during op bootstrapping at decorator time"""
        op = self._ops.pop(name)
        op.name = new_name
        self._ops[new_name] = op

        self._ops_by_common_name[op.common_name].pop(name)
        self._ops_by_common_name.setdefault(op.common_name, {})[new_name] = op

        old_version = op.version

        # TODO(DG): find a better way to do this than to save the op again
        # see comment here: https://github.com/wandb/weave-internal/pull/554#discussion_r1103875156
        if op.location is not None:
            ref = storage.save(op, name=new_name)
            location = ref.artifact.uri_obj
            version = ref.version
            op.version = version
            op.location = location

        if op.version is not None:
            self._op_versions.pop((name, old_version))
            self._op_versions[(new_name, op.version)] = op

    # def register_type(self, type: weave_types.Type):
    #    self._types[type.name] = type

    # def get_type(self, name):
    #    # TODO, don't repeat this all the time
    #    types = all_subclasses(weave_types.Type)
    #    for type in types:
    #        if type.name == name:
    #            return type
    #    raise Exception("type not found")


# Processes have a singleton MemoryRegistry
memory_registry = Registry()



from ..panel import Panel

# special
from .panel_expression import *
from .panel_auto import *

# layout
from .panel_labeled_item import LabeledItem
from .panel_card import Card, CardTab
from .panel_group import Group, GroupLayoutFlow, GroupPanel
from .panel_each import Each

from .panel_facet_tabs import FacetTabs
from .panel_sections import Sections

# navigation
from .panel_weavelink import WeaveLink

from .panel_table import Table, TableColumn, ColumnDef
from .panel_plot import Plot, Series, PlotConstants
from .panel_facet import Facet
from .panel_each_column import EachColumn
from .panel_color import Color
from .panel_html import PanelHtml
from .panel_markdown import PanelMarkdown
from .panel_trace import Trace


# sidebar specific
from .panel_query import Query, QueryConfig, QueryCondition

# editors
from .panel_slider import Slider, SliderConfig
from .panel_select import SelectEditor, SelectEditorConfig
from .panel_daterange import DateRange
from .panel_string_editor import StringEditor
from .panel_function_editor import FunctionEditor, FunctionEditorConfig

# Non-standard editor (todo: update)
from .panel_object_picker import ObjectPicker, ObjectPickerConfig

# basic
from .panel_basic import *

# domain
from .panel_domain import *

# Incomplete
from .panel_histogram import *

# legacy
from .panel_legacy import *

# top level board
from .panel_board import Board, BoardPanel, BoardPanelLayout



import re
import typing
from ..compile_domain import wb_gql_op_plugin
from ..api import op
from . import wb_domain_types as wdt
from .wandb_domain_gql import (
    gql_prop_op,
    gql_direct_edge_op,
)
from .. import weave_types as types
import urllib
from .. import errors

# Section 1/6: Tag Getters
#

# Section 2/6: Root Ops
@op(
    name="root-allReportsGQLResolver",
    input_type={"gql_result": types.TypedDict({})},
    output_type=types.List(wdt.ReportType),
    hidden=True,
)
def root_all_reports_gql_resolver(gql_result):
    return [
        wdt.Report.from_gql(report["node"])
        for report in gql_result["instance"]["views_500"]["edges"]
        if report["node"]["type"] == "runs"  # yes, this is how we filter to Reports!?
    ]


@op(
    name="root-allReports",
    input_type={},
    output_type=types.List(wdt.ReportType),
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: f"""
        instance {{
            views_500: views(limit: 500) {{
                edges {{
                    node {{
                        type
                        {wdt.Report.REQUIRED_FRAGMENT}
                        {inner}
                    }}
                }}
            }}
        }}""",
        is_root=True,
        root_resolver=root_all_reports_gql_resolver,
    ),
)
def root_all_reports():
    raise errors.WeaveGQLCompileError(
        "root-allReports should not be executed directly. If you see this error, it is a bug in the Weave compiler."
    )


# Section 3/6: Attribute Getters
gql_prop_op(
    "report-createdAt",
    wdt.ReportType,
    "createdAt",
    types.Timestamp(),
)
gql_prop_op(
    "report-viewcount",
    wdt.ReportType,
    "viewCount",
    types.Number(),
)

# Section 4/6: Direct Relationship Ops
gql_direct_edge_op(
    "report-project",
    wdt.ReportType,
    "project",
    wdt.ProjectType,
)

gql_direct_edge_op(
    "report-creator",
    wdt.ReportType,
    "user",
    wdt.UserType,
)

# Section 5/6: Connection Ops
#

# Section 6/6: Non Standard Business Logic Ops

# Logic taken exactly from Weave0 / frontend
def make_name_and_id(id: str, name: typing.Optional[str]) -> str:
    id = id.replace("=", "")
    if name is None:
        return id
    name = re.sub(r"\W", "-", name)
    name = re.sub(r"-+", "-", name)
    name = urllib.parse.quote(name)

    return f"{name}--{id}"


@op(
    name="report-link",
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: f"""
    id
    displayName
    project {{
        id
        name 
        entity {{
            id
            name
        }}
    }}
"""
    ),
)
def link(report: wdt.Report) -> wdt.Link:
    project = report.gql["project"]
    entity = project["entity"]

    project_name = project["name"]
    entity_name = entity["name"]
    report_id = report.gql["id"]
    report_name = report.gql["displayName"]

    url = f"/{entity_name}/{project_name}/reports/{make_name_and_id(report_id, report_name)}"
    return wdt.Link(report_name, url)



import PIL
import os
import typing

import weave


class OxfordIIITPetDatasetItem(typing.TypedDict):
    id: str
    class_id: int
    species: str
    breed: str
    image: PIL.Image.Image


@weave.type()
class PetadataRenderConfig:
    pass


SPECIES = ["cat", "dog"]

CAT_BREEDS = [
    "Abyssinian",
    "Bengal",
    "Birman",
    "Bombay",
    "British_Shorthair",
    "Egyptian_Mau",
    "Maine_Coon",
    "Persian",
    "Ragdoll",
    "Russian_Blue",
    "Siamese",
    "Sphynx",
]

DOG_BREEDS = [
    "american_bulldog",
    "american_pit_bull_terrier",
    "basset_hound",
    "beagle",
    "boxer",
    "chihuahua",
    "english_cocker_spaniel",
    "english_setter",
    "german_shorthaired",
    "great_pyrenees",
    "havanese",
    "japanese_chin",
    "keeshond",
    "leonberger",
    "miniature_pinscher",
    "newfoundland",
    "pomeranian",
    "pug",
    "saint_bernard",
    "samoyed",
    "scottish_terrier",
    "shiba_inu",
    "staffordshire_bull_terrier",
    "wheaten_terrier",
    "yorkshire_terrier",
]


@weave.type()
class PetDatasetPanel(weave.Panel):
    id = "PetDatasetPanel"
    input_node: weave.Node[list[OxfordIIITPetDatasetItem]]

    @weave.op()
    def render(self) -> weave.panels.Table:
        return weave.panels.Table(
            self.input_node,
            columns=[
                lambda item: item["id"],
                lambda item: item["image"],
                lambda item: item["combined_class_id"],
                lambda item: item["species_id"],
                lambda item: item["breed_id"],
            ],
        )


@weave.op()
def petdataset(
    raw_data_path: str,
) -> list[OxfordIIITPetDatasetItem]:  # TODO: use weave path
    # Download from here: https://www.robots.ox.ac.uk/~vgg/data/pets/

    annotations_list_path = os.path.join(raw_data_path, "annotations", "list.txt")
    items: list[OxfordIIITPetDatasetItem] = []

    # For now, shuffle them so we get an interesting subset
    import random

    lines = open(annotations_list_path).read().split("\n")
    random.shuffle(lines)

    with open(annotations_list_path) as f:
        for i, line in zip(range(50), lines):
            # skip comments
            if line.startswith("#"):
                continue
            img_id, class_id, species_id, breed_id = line.strip().split(" ")
            img_path = os.path.join(raw_data_path, "images", f"{img_id}.jpg")
            img = PIL.Image.open(img_path)
            img.load()

            if species_id == "1":
                species = "cat"
                breed = CAT_BREEDS[int(breed_id) - 1]
            else:
                species = "dog"
                breed = DOG_BREEDS[int(breed_id) - 1]

            items.append(
                OxfordIIITPetDatasetItem(
                    id=img_id,
                    class_id=int(class_id),
                    species=species,
                    breed=breed,
                    image=img,
                )
            )
    return items



import dataclasses
import typing

import weave
from .. import panel
from .. import weave_internal

from . import panel_group


# @dataclasses.dataclass
# class Board:
#     vars: typing.Dict[str, typing.Any]
#     panels: typing.Dict[str, typing.Any]


@dataclasses.dataclass
class BoardPanelLayout:
    x: int
    y: int
    h: int
    w: int


@dataclasses.dataclass
class BoardPanel:
    panel: typing.Any
    id: typing.Optional[str] = None
    layout: typing.Optional[BoardPanelLayout] = None


def Board(vars, panels: list[BoardPanel], editable=True):
    main_items = {}
    main_panel_layouts: list[panel_group.LayedOutPanel] = []
    for i, p in enumerate(panels):
        id = p.id
        if id is None:
            id = "panel" + str(i)
        main_items[id] = p.panel
        if p.layout is not None:
            main_panel_layouts.append(
                panel_group.LayedOutPanel(
                    id=id,
                    layout=panel_group.LayoutParameters(
                        x=p.layout.x,
                        y=p.layout.y,
                        w=p.layout.w,
                        h=p.layout.h,
                    ),
                )
            )

    showExpressions = True if editable else "titleBar"
    return panel_group.Group(
        config=panel_group.GroupConfig(
            layoutMode="horizontal",
            showExpressions=False,
        ),
        items={
            "sidebar": panel_group.Group(
                config=panel_group.GroupConfig(
                    layoutMode="vertical",
                    equalSize=False,
                    style="width:300px;",
                    showExpressions=showExpressions,  # type: ignore
                    allowedPanels=[
                        "Expression",
                        "Query",
                        "Slider",
                        "StringEditor",
                        "SelectEditor",
                        "DateRange",
                    ],
                    enableAddPanel=True,
                    childNameBase="var",
                ),
                items=vars,
            ),
            "main": panel_group.Group(
                config=panel_group.GroupConfig(
                    layoutMode="grid",
                    showExpressions=showExpressions,  # type: ignore
                    enableAddPanel=True,
                    gridConfig=panel_group.PanelBankSectionConfig(
                        id="grid0",
                        name="Section 0",
                        panels=main_panel_layouts,
                        isOpen=True,
                        flowConfig=panel_group.PanelBankFlowSectionConfig(
                            snapToColumns=True,
                            columnsPerPage=3,
                            rowsPerPage=2,
                            gutterWidth=10,
                            boxWidth=256,
                            boxHeight=256,
                        ),
                        type="grid",
                        sorted=0,
                    ),
                ),
                items=main_items,
            ),
        },
    )



# Implements backward compatibility for existing W&B Media types.

import dataclasses
import typing
from ..language_features.tagging.tag_store import isolated_tagging_context
from .. import types
from .. import errors
from .. import api as weave
from .. import artifact_fs
from ..ops_primitives import html
from ..ops_primitives import markdown


@dataclasses.dataclass(frozen=True)
class LegacyImageArtifactFileRefType(types.ObjectType):
    name = "legacy-image-file"

    def property_types(self) -> dict[str, types.Type]:
        raise errors.WeaveTypeError(
            f"LegacyImageArtifactFileRefType should never be used! it is a temp type"
        )

    @classmethod
    def type_of_instance(cls, obj):
        raise errors.WeaveTypeError(
            f"LegacyImageArtifactFileRefType should never be used! it is a temp type"
        )

    def _to_dict(self) -> dict:
        raise errors.WeaveTypeError(
            f"LegacyImageArtifactFileRefType should never be used! it is a temp type"
        )

    @classmethod
    def from_dict(cls, d):
        raise errors.WeaveTypeError(
            f"LegacyImageArtifactFileRefType should never be used! it is a temp type"
        )


@dataclasses.dataclass(frozen=True)
class ImageArtifactFileRefType(types.ObjectType):
    name = "image-file"
    boxLayers: typing.Union[types.Type, dict] = types.TypedDict({})
    boxScoreKeys: typing.Union[types.Type, list] = types.List(types.UnknownType())
    maskLayers: typing.Union[types.Type, dict] = types.TypedDict({})
    classMap: typing.Union[types.Type, dict] = types.TypedDict({})

    # TODO: This should probably be standard for Type!
    def __post_init__(self):
        for type_attr in self.type_attrs():
            self.__dict__[type_attr] = types.parse_constliteral_type(
                self.__dict__[type_attr]
            )

    def _to_dict(self) -> dict:
        d: dict = {"_is_object": True}
        d = self.class_to_dict()
        d["_is_object"] = True
        d["boxLayers"] = types.constliteral_type_to_json(self.boxLayers)  # type: ignore
        d["boxScoreKeys"] = types.constliteral_type_to_json(self.boxScoreKeys)  # type: ignore
        d["maskLayers"] = types.constliteral_type_to_json(self.maskLayers)  # type: ignore
        d["classMap"] = types.constliteral_type_to_json(self.classMap)  # type: ignore
        return d

    # TODO: This should probably be standard for Type!
    @classmethod
    def from_dict(cls, d):
        return cls(
            d.get("boxLayers", {}),
            d.get("boxScoreKeys", []),
            d.get("maskLayers", {}),
            d.get("classMap", {}),
        )

    def property_types(self) -> dict[str, types.Type]:
        boxLayers = types.constliteral_type_to_json(self.boxLayers)  # type: ignore
        boxScoreKeys = types.constliteral_type_to_json(self.boxScoreKeys)  # type: ignore
        maskLayers = types.constliteral_type_to_json(self.maskLayers)  #   type: ignore
        res = {
            "artifact": artifact_fs.FilesystemArtifactType(),
            "path": types.String(),
            "format": types.String(),
            "height": types.Int(),
            "width": types.Int(),
            "sha256": types.String(),
            "boxes": types.optional(
                types.TypedDict(
                    {
                        box_key: types.optional(
                            types.List(
                                types.TypedDict(
                                    {
                                        "box_caption": types.optional(types.String()),
                                        "class_id": types.Int(),
                                        "domain": types.optional(types.String()),
                                        # The framework should do this merge types for us. But for
                                        # now we have to do it manually.
                                        "position": types.merge_types(
                                            types.TypedDict(
                                                {
                                                    "maxX": types.Float(),
                                                    "maxY": types.Float(),
                                                    "minX": types.Float(),
                                                    "minY": types.Float(),
                                                }
                                            ),
                                            types.TypedDict(
                                                {
                                                    "height": types.Float(),
                                                    "middle": types.List(types.Float()),
                                                    "width": types.Float(),
                                                }
                                            ),
                                        ),
                                        "scores": types.optional(
                                            types.TypedDict(
                                                {
                                                    score_key: types.optional(
                                                        types.Float()
                                                    )
                                                    for score_key in boxScoreKeys
                                                }
                                            )
                                        ),
                                    }
                                )
                            )
                        )
                        for box_key in boxLayers.keys()
                    }
                )
            ),
            "masks": types.optional(
                types.TypedDict(
                    {
                        mask_key: types.optional(
                            types.TypedDict(
                                {
                                    "_type": types.String(),
                                    "path": types.String(),
                                    "sha256": types.optional(types.String()),
                                }
                            )
                        )
                        for mask_key in maskLayers.keys()
                    }
                )
            ),
        }
        return res

    @classmethod
    def type_of_instance(cls, obj):
        if obj.boxes:
            boxLayers = {
                key: [row["class_id"] for row in value]
                for key, value in obj.boxes.items()
                if value is not None
            }
        else:
            boxLayers = {}
        boxScoreKeysSet = {}
        for boxLayer in boxLayers:
            for box in obj.boxes[boxLayer]:
                if box["scores"] is not None:
                    boxScoreKeysSet.update(dict.fromkeys(box["scores"].keys()))
        boxScoreKeys = list(boxScoreKeysSet)

        if obj.masks:
            maskLayers = {
                # Empty array here. We really should read in the mask data, then
                # find all the unique values. but that is really costly and is not
                # worth it. This is because `type_of_instance` is only called when
                # we are trying to re-serialize for transmission back to the client.
                # In these cases, having the correct classes in each key is not
                # needed.
                key: []
                for key in obj.masks.keys()
            }
        else:
            maskLayers = {}
        return cls(
            boxLayers=boxLayers,
            boxScoreKeys=boxScoreKeys,
            maskLayers=maskLayers,
            classMap={},
        )


@weave.weave_class(weave_type=ImageArtifactFileRefType)
@dataclasses.dataclass
class ImageArtifactFileRef:
    artifact: artifact_fs.FilesystemArtifact
    path: str
    format: str
    height: int
    width: int
    sha256: str
    boxes: typing.Optional[dict[str, list[dict]]] = dataclasses.field(
        default_factory=dict
    )  # type: ignore
    masks: typing.Optional[dict[str, dict[str, str]]] = dataclasses.field(
        default_factory=dict
    )  # type: ignore
    classes: typing.Optional[typing.Optional[dict]] = None


ImageArtifactFileRef.WeaveType = ImageArtifactFileRefType  # type: ignore
ImageArtifactFileRefType.instance_class = ImageArtifactFileRef
ImageArtifactFileRefType.instance_classes = ImageArtifactFileRef


@weave.type(__override_name="audio-file")  # type: ignore
class AudioArtifactFileRef:
    artifact: artifact_fs.FilesystemArtifact
    path: str
    sha256: str


@weave.type(__override_name="bokeh-file")  # type: ignore
class BokehArtifactFileRef:
    artifact: artifact_fs.FilesystemArtifact
    path: str
    sha256: str


@weave.type(__override_name="video-file")  # type: ignore
class VideoArtifactFileRef:
    artifact: artifact_fs.FilesystemArtifact
    path: str
    sha256: str


@weave.type(__override_name="object3D-file")  # type: ignore
class Object3DArtifactFileRef:
    artifact: artifact_fs.FilesystemArtifact
    path: str
    sha256: str


@weave.type(__override_name="molecule-file")  # type: ignore
class MoleculeArtifactFileRef:
    artifact: artifact_fs.FilesystemArtifact
    path: str
    sha256: str


@weave.type(__override_name="html-file")  # type: ignore
class HtmlArtifactFileRef:
    artifact: artifact_fs.FilesystemArtifact
    path: str
    sha256: str


# When a WB table is written to disk, it accumulates all the NDArrays into a
# single file so that these columns can be stored more efficiently. In the UI,
# we completely ignore such columns. The underlying data is even Nulled out!
#  With Weave1, we actually have the possibility of doing something with them if
# we wanted to! There are additional properties which we could use to add value:
# {"params": {"serialization_path": {"key": "output", "path":
# "media/serialized_data/498587e8.npz"}, "shape": [10]} However, I suspect we
# will just use the newer Weave1 NDArray type instead.

# 3/22/23: However as of this comment, we just return NoneType() instead of this
# type when we encounter legacy ndarray, since the table data is None
class LegacyTableNDArrayType(types.Type):
    name = "ndarray"


# This shows a pattern for how to convert an in memory object (Html)
# to a W&B media type style FileRef, so that the existing frontend
# code can work with it.
@weave.op()
def html_file(html: html.Html) -> HtmlArtifactFileRef:
    from weave import storage

    # This is a ref to the html object
    with isolated_tagging_context():
        ref = storage.save(html)
    path = ref.path
    if path is None:
        raise errors.WeaveInternalError("storage save returned None path")
    file_path = path + ".html"
    return HtmlArtifactFileRef(ref.artifact, file_path, file_path)


# Yet another pattern for returning a file inside an artifact!
# In this case, the WeaveJS Markdown panel expects a 'file' type
# (with extension in the type).
# TODO: merge all these patterns!!!!
@weave.op(
    output_type=artifact_fs.FilesystemArtifactFileType(
        weave.types.Const(weave.types.String(), "md")  # type: ignore
    )
)
def markdown_file(md: markdown.Markdown):
    from weave import storage

    with isolated_tagging_context():
        ref = storage.save(md)
    path = ref.path
    if path is None:
        raise errors.WeaveInternalError("storage save returned None path")
    return artifact_fs.FilesystemArtifactFile(ref.artifact, path + ".md")


ArtifactAssetType = types.union(
    ImageArtifactFileRef.WeaveType(),  # type: ignore
    AudioArtifactFileRef.WeaveType(),  # type: ignore
    BokehArtifactFileRef.WeaveType(),  # type: ignore
    VideoArtifactFileRef.WeaveType(),  # type: ignore
    Object3DArtifactFileRef.WeaveType(),  # type: ignore
    MoleculeArtifactFileRef.WeaveType(),  # type: ignore
    HtmlArtifactFileRef.WeaveType(),  # type: ignore
)


@weave.op(
    name="asset-artifactVersion",
    input_type={"asset": ArtifactAssetType},
    output_type=artifact_fs.FilesystemArtifactType(),
)
def artifactVersion(asset):
    return asset.artifact


@weave.op(
    name="asset-file",
    input_type={"asset": ArtifactAssetType},
    output_type=artifact_fs.FilesystemArtifactFileType(),
)
def asset_file(asset):
    return asset.artifact.path_info(asset.path)



import typing
import json
import openai
import os
import tempfile
import time

import weave

openai.api_key_path = os.path.expanduser("~/.openai-apikey")


class StoredFileType(weave.types.ObjectType):
    name = "openai_stored_file"
    purpose: str

    def __init__(self, purpose=weave.types.String()):
        self.purpose = purpose

    def property_types(self):
        return {
            "bytes": weave.types.Int(),
            "created_at": weave.types.Int(),  # TODO: convert to date
            "filename": weave.types.String(),
            "id": weave.types.String(),
            "object": weave.types.String(),
            "purpose": self.purpose,
            "status": weave.types.String(),  # TODO: enum?
            "status_details": weave.types.none_type,
        }


# TODO: This is a mirror of OpenAI.File, but would we just attach a type
#     to that and not need to make this class at all?


class StoredFile:
    def __init__(
        self, bytes, created_at, filename, id, object, purpose, status, status_details
    ):
        self.bytes = bytes
        self.created_at = created_at
        self.filename = filename
        self.id = id
        self.object = object
        self.purpose = purpose
        self.status = status
        self.status_details = status_details

    # TODO: currently unused
    def get_openai_file(self):
        return openai.File.retrieve(id=self.id)

    @weave.op(
        input_type={
            "self": StoredFileType(weave.types.String()),
        },
        output_type=weave.types.String(),
    )
    def contents(self):
        return openai.File.download(id=self.id).decode()


StoredFileType.instance_classes = StoredFile
StoredFileType.instance_class = StoredFile


class Gpt3DatasetType(StoredFileType):
    name = "gpt3_dataset"

    def __init__(self):
        self.purpose = weave.types.Const(weave.types.String(), "fine-tune")


# TODO: we can make a storage manager for OpenAI Files that
#     allows transactional editing of OpenAI files from panels


@weave.weave_class(weave_type=Gpt3DatasetType)
class Gpt3Dataset(StoredFile):
    @weave.op(
        input_type={
            "self": StoredFileType(),
        },
        output_type=weave.types.List(
            weave.types.TypedDict(
                {"prompt": weave.types.String(), "completion": weave.types.String()}
            )
        ),
    )
    def items(self):
        items = []
        for line in openai.File.download(id=self.id).strip().split(b"\n"):
            items.append(json.loads(line))
        return items


Gpt3DatasetType.instance_classes = Gpt3Dataset
Gpt3DatasetType.instance_class = Gpt3Dataset


class Gpt3FineTuneResultsType(StoredFileType):
    name = "gpt3_fine_tune_results_type"

    def __init__(self):
        self.purpose = weave.types.Const(weave.types.String(), "fine-tune-results")


# TODO: we can make a storage manager for OpenAI Files that
#     allows transactional editing of OpenAI files from panels


class Gpt3FineTuneResults(StoredFile):
    @weave.op(
        input_type={
            "self": StoredFileType(),
        },
        output_type=weave.types.List(
            weave.types.TypedDict(
                {
                    "step": weave.types.Int(),
                    "elapsed_tokens": weave.types.Float(),
                    "elapsed_examples": weave.types.Float(),
                    "training_loss": weave.types.Float(),
                    "training_sequence_accuracy": weave.types.Float(),
                    "training_token_accuracy": weave.types.Float(),
                }
            )
        ),
    )
    def table(self):
        contents = openai.File.download(id=self.id)
        with tempfile.NamedTemporaryFile() as f:
            f.write(contents)
            f.seek(0)
            return weave.ops.csv_.load_csv(f.name)


Gpt3FineTuneResultsType.instance_classes = Gpt3FineTuneResults
Gpt3FineTuneResultsType.instance_class = Gpt3FineTuneResults


@weave.op(
    name="openai-uploadgpt3dataset",
    input_type={
        "items": weave.types.List(
            weave.types.TypedDict(
                {"prompt": weave.types.String(), "completion": weave.types.String()}
            )
        )
    },
    output_type=Gpt3DatasetType(),
    # TODO: we shouldn't need this!
    render_info={"type": "function"},
)
def upload_gpt3_dataset(items):
    with tempfile.TemporaryFile(mode="w+") as f:
        for item in items:
            json.dump(dict(item), f)
            f.write("\n")
        f.seek(0)
        # TODO: purpose can be other stuff, its an enum
        resp = openai.File.create(file=f, purpose="fine-tune")
    return Gpt3Dataset(**resp)


class Gpt3ModelType(weave.types.ObjectType):
    name = "gpt3-model"

    def __init__(self):
        pass

    def property_types(self):
        return {"id": weave.types.String(), "version": weave.types.String()}


class Gpt3ModelCompleteReturnChoices(typing.TypedDict):
    text: str
    index: int
    logprobs: None
    finish_reason: str


class Gpt3ModelCompleteReturn(typing.TypedDict):
    id: str
    object: str
    created: int
    model: str
    choices: list[Gpt3ModelCompleteReturnChoices]


@weave.weave_class(weave_type=Gpt3ModelType)
class Gpt3Model:
    def __init__(self, id, version):
        self.id = id
        # self.version is a HACK to force a new object for the demo
        # TDOO
        self.version = version

    @weave.op(
        name="gpt3model-complete",
    )
    def complete(self, prompt: str) -> Gpt3ModelCompleteReturn:
        sleep = 1
        completion_params = {
            "model": self.id,
            "prompt": prompt,
            "max_tokens": 256,
            "n": 3,
            "temperature": 0.9,
        }
        for _ in range(5):
            try:
                return openai.Completion.create(**completion_params)
            except openai.error.RateLimitError:
                # This error occurs if a model is newly trained or hasn't
                # been queried for awhile, while the OpenAI backend loads
                # it up.
                time.sleep(sleep)
                sleep *= 2
        return openai.Completion.create(**completion_params)


Gpt3ModelType.instance_classes = Gpt3Model
Gpt3ModelType.instance_class = Gpt3Model


class Gpt3FineTuneType(weave.types.ObjectType):
    name = "gpt3_fine_tune_type"

    def __init__(self):
        pass

    def property_types(self):
        return {
            "id": weave.types.String(),
            "status": weave.types.String(),
            "fine_tuned_model": weave.types.optional(weave.types.String()),
            # TODO: This is plural in the OpenAI API. Should we make it a list?
            "result_file": weave.types.optional(Gpt3FineTuneResultsType()),
        }


@weave.weave_class(weave_type=Gpt3FineTuneType)
class Gpt3FineTune:
    def __init__(self, id, status, fine_tuned_model, result_file):
        self.id = id
        self.status = status
        self.fine_tuned_model = fine_tuned_model
        self.result_file = result_file

    def update(self):
        fine_tune = openai.FineTune.retrieve(id=self.id)
        # print("FINE_TUNE", fine_tune)
        self.status = fine_tune["status"]
        self.fine_tuned_model = fine_tune["fine_tuned_model"]
        self.result_file = None
        if fine_tune["result_files"]:
            self.result_file = Gpt3FineTuneResults(**fine_tune["result_files"][0])

    @weave.op(
        name="gpt3finetune-model",
        input_type={
            "self": Gpt3FineTuneType(),
        },
        output_type=Gpt3ModelType(),
    )
    def model(self):
        return Gpt3Model(self.fine_tuned_model, self.id)


Gpt3FineTuneType.instance_classes = Gpt3FineTune
Gpt3FineTuneType.instance_class = Gpt3FineTune


@weave.op(
    render_info={"type": "function"},
    input_type={
        "training_dataset": weave.types.List(
            weave.types.TypedDict(
                {"prompt": weave.types.String(), "completion": weave.types.String()}
            )
        ),
        "hyperparameters": weave.types.TypedDict(
            {
                "n_epochs": weave.types.Int(),
                # "learning_rate_multiplier": weave.types.Float(),
            }
        ),
    },
    output_type=weave.types.RunType(
        output=Gpt3FineTuneType(),
    ),
)
def finetune_gpt3(training_dataset, hyperparameters, _run=None):
    uploaded = weave.use(upload_gpt3_dataset(training_dataset))
    create_args = {"training_file": uploaded.id}
    for k, v in hyperparameters.items():
        create_args[k] = v
    weave.use(_run.print_("Creating fine tune"))
    resp = openai.FineTune.create(**create_args)
    fine_tune = Gpt3FineTune(
        id=resp["id"],
        status=resp["status"],
        fine_tuned_model=resp["fine_tuned_model"],
        result_file=None,
    )
    while True:
        fine_tune.update()
        time.sleep(3)
        weave.use(_run.print_("Fine_tune status: %s" % fine_tune.status))
        if fine_tune.status == "succeeded":
            break
    weave.use(_run.set_output(fine_tune))


@weave.op(
    render_info={"type": "function"},
    input_type={
        "training_dataset": weave.types.List(
            weave.types.TypedDict(
                {"prompt": weave.types.String(), "completion": weave.types.String()}
            )
        ),
        "hyperparameters": weave.types.TypedDict(
            {
                "n_epochs": weave.types.Int(),
                # "learning_rate_multiplier": weave.types.Float(),
            }
        ),
    },
    output_type=weave.types.RunType(
        output=Gpt3FineTuneType(),
    ),
)
def finetune_gpt3_demo(training_dataset, hyperparameters, _run=None):
    print("!!! FINE TUNE DEMO, NOT REALLY FINE-TUNING !!!")
    weave.use(_run.print_("Creating fine tune"))
    resp = {
        "id": str(training_dataset._ref) + str(json.dumps(hyperparameters)),
        "status": "running",
        "fine_tuned_model": None,
    }
    fine_tune = Gpt3FineTune(
        id=resp["id"],
        status=resp["status"],
        fine_tuned_model=resp["fine_tuned_model"],
        result_file=None,
    )
    i = 0
    while True:
        weave.use(_run.print_("Fine_tune status: running"))
        i += 1
        if i == 6:
            fine_tune.fine_tuned_model = "ada:ft-wandb-2021-10-05-23-25-22"
            fine_tune.status = "succeeded"
            weave.use(_run.print_("Fine_tune status: succeeded"))
            break
        time.sleep(3)
    weave.use(_run.set_output(fine_tune))


@weave.op()
def gpt3_davinci_2() -> Gpt3Model:
    return Gpt3Model("text-davinci-002", "text-davinci-002")


# A very simple predict API. Should probably return more than one
# result though.
@weave.op()
def gpt3_predict(prompt: str) -> str:
    completion_params = {
        "model": "text-davinci-002",
        "prompt": prompt,
        "max_tokens": 256,
        "n": 3,
        "temperature": 0.9,
    }
    sleep = 1

    for _ in range(5):
        try:
            result = openai.Completion.create(**completion_params)
            return result["choices"][0]["text"]
        except openai.error.RateLimitError:
            # This error occurs if a model is newly trained or hasn't
            # been queried for awhile, while the OpenAI backend loads
            # it up.
            time.sleep(sleep)
            sleep *= 2
    result = openai.Completion.create(**completion_params)["choices"][0]["text"]

    return result



import inspect
import typing
import dataclasses

from . import weave_types as types
from . import infer_types
from . import decorator_class
from . import errors
from . import decorator_op

_py_type = type


def type(
    __override_name: typing.Optional[str] = None,
    __is_simple: bool = False,
    __init: typing.Optional[bool] = None,
):
    def wrap(target):
        init = False
        if __init is not None:
            init = __init
        else:
            init = getattr(target.__bases__[0], "__init", True)
        target.__init = init
        dc = dataclasses.dataclass(target, init=init)
        fields = dataclasses.fields(dc)
        target_name = target.__name__
        if __override_name is not None:
            target_name = __override_name

        base_type = types.ObjectType
        if target.__bases__:
            # Add the first base classes as the type base.
            # TODO: should we add all bases?
            target_base0 = target.__bases__[0]
            if hasattr(target_base0, "WeaveType"):
                base_type = target_base0.WeaveType

        if __is_simple:
            bases = (
                base_type,
                types._PlainStringNamedType,
            )
        else:
            bases = (base_type,)
        TargetType = _py_type(f"{target_name}Type", bases, {})
        TargetType.name = target_name
        TargetType.instance_classes = target
        TargetType.instance_class = target

        type_vars: dict[str, types.Type] = {}
        static_property_types: dict[str, types.Type] = {}
        for field in fields:
            if isinstance(field.type, typing.TypeVar):
                # This is a Python type variable
                type_vars[field.name] = types.Any()
            else:
                try:
                    weave_type = infer_types.python_type_to_type(field.type)
                except TypeError:
                    # hmmm... Exception rewriting. Am I OK with this? Could be overly aggressive.
                    # TODO: decide if we should do this
                    raise errors.WeaveDefinitionError(
                        f"{target}.{field.name} is not a valid python type (a class or type)"
                    )

                if types.type_is_variable(weave_type):
                    # this is a Weave type with a type variable in it
                    type_vars[field.name] = weave_type
                else:
                    static_property_types[field.name] = weave_type

        if type_vars:
            setattr(TargetType, "__annotations__", {})
        for name, default_type in type_vars.items():
            setattr(TargetType, name, default_type)
            TargetType.__dict__["__annotations__"][name] = types.Type

        def property_types_method(self):
            property_types = {}
            for name in type_vars:
                property_types[name] = getattr(self, name)
            for name, prop_type in static_property_types.items():
                property_types[name] = prop_type
            return property_types

        TargetType.property_types = property_types_method

        if not type_vars:
            # workaround rich bug. dataclasses __repr__ doesn't print if the dataclass has no
            # fields. Insert our own __repr__ in that case so rich doesn't try to do its
            # special dataclass repr.
            TargetType.__repr__ = lambda self: f"{self.__class__.__name__}()"

        TargetType = dataclasses.dataclass(frozen=True)(TargetType)

        dc.WeaveType = TargetType
        decorator_class.weave_class(weave_type=TargetType)(dc)

        # constructor op for this type. due to a circular dependency with ArrowWeave* types, we
        # define the vectorized constructor ops in vectorize.py instead of here
        @decorator_op.op(
            name=f"objectConstructor-_new_{target_name.replace('-', '_')}",
            input_type={
                field.name: static_property_types.get(field.name, None)
                or type_vars[field.name]
                for field in fields
            },
            output_type=TargetType(),
            render_info={"type": "function"},
        )
        def constructor(**attributes):
            return dc(
                **{field.name: attributes[field.name] for field in fields if field.init}
            )

        dc.constructor = constructor

        return dc

    return wrap



import json
import typing
from ...compile_domain import wb_gql_op_plugin
from ...api import op
from ... import weave_types as types
from .. import wb_domain_types as wdt
from ..wandb_domain_gql import (
    _make_alias,
)
from .. import wb_util
from ... import engine_trace


import pyarrow as pa

from . import history_op_common


tracer = engine_trace.tracer()


@op(
    name="run-refine_history_type",
    render_info={"type": "function"},
    hidden=True,
    plugins=wb_gql_op_plugin(lambda inputs, inner: "historyKeys"),
)
def refine_history_type(run: wdt.Run) -> types.Type:
    return types.List(history_op_common.refine_history_type(run))


@op(
    name="run-history",
    refine_output_type=refine_history_type,
    plugins=wb_gql_op_plugin(history_op_common.make_run_history_gql_field),
    output_type=types.List(types.TypedDict({})),
)
def history(run: wdt.Run):
    # We return mock data here since we will be replaced with the `_with_columns`
    # version in a compile pass if specific columns are needed
    return history_op_common.mock_history_rows(run, False)


@op(
    render_info={"type": "function"},
    plugins=wb_gql_op_plugin(lambda inputs, inner: "historyKeys"),
    hidden=True,
)
def refine_history_with_columns_type(
    run: wdt.Run, history_cols: list[str]
) -> types.Type:
    return types.List(
        history_op_common.refine_history_type(
            run, columns=history_op_common.get_full_columns_prefixed(run, history_cols)
        )
    )


@op(
    name="run-history_with_columns",
    refine_output_type=refine_history_with_columns_type,
    plugins=wb_gql_op_plugin(history_op_common.make_run_history_gql_field),
    output_type=types.List(types.TypedDict({})),
    hidden=True,
)
def history_with_columns(run: wdt.Run, history_cols: list[str]):
    return _get_history(
        run, history_op_common.get_full_columns_prefixed(run, history_cols)
    )


def _get_history(run: wdt.Run, columns=None):
    with tracer.trace("get_history") as span:
        span.set_tag("history_version", 1)
    # we have fetched some specific rows.
    # download the files from the urls

    with tracer.trace("read_history_parquet"):
        parquet_history = history_op_common.read_history_parquet(
            run, columns=columns
        ) or pa.table([])

    # turn the liveset into an arrow table. the liveset is a list of dictionaries
    live_data = run.gql["sampledParquetHistory"]["liveData"]

    with tracer.trace("liveSet.impute"):
        for row in live_data:
            for colname in columns:
                if colname not in row:
                    row[colname] = None

    # get binary fields from history schema - these are serialized json
    binary_fields = [
        field.name for field in parquet_history.schema if pa.types.is_binary(field.type)
    ]

    with tracer.trace("pq.to_pylist"):
        parquet_history = parquet_history.to_pylist()

    # deserialize json
    with tracer.trace("json.loads"):
        for field in binary_fields:
            for row in parquet_history:
                if row[field] is not None:
                    row[field] = json.loads(row[field])

    # parquet stores step as a float, but we want it as an int
    for row in parquet_history:
        row["_step"] = int(row["_step"])

    history = parquet_history + live_data

    with tracer.trace("process_run_dict_obj"):
        return [
            wb_util.process_run_dict_obj(
                row,
                wb_util.RunPath(
                    run.gql["project"]["entity"]["name"],
                    run.gql["project"]["name"],
                    run.gql["name"],
                ),
            )
            for row in history
        ]



import json
from dataclasses import field, dataclass
import typing
from ..decorator_type import type as weave_type
from .. import weave_types as types

"""
This file contains all the "W&B Domain Types". Each domain type should
correspond to exactly 1 type/interface in the WB GQL schema. Not all types are
represented as we only have implemented the subset that is required for Weave1
compatibility. Each type follows the same pattern for declaration. Consider
User:

@weave_type("user", True)                       # Sets the Weave type to be a plain string called "user"
class User(GQLTypeMixin):                       # Creates the Weave type in Weave1 and adds the GQLTypeMixin
    REQUIRED_FRAGMENT = "id name"               # Declares a minimum GQL fragment that should be used whenever loading this type from GQL

UserType = User.WeaveType()  # type: ignore     # Exports a casted singleton type for easier construction in domain ops.
UserType = typing.cast(types.Type, UserType) .  # Note: the type: ignore and cast could be removed if we fix the .pyi file for decorator_type

There are two main concepts to grasp:
    1. The `REQUIRED_FRAGMENT` is used to specify the minimum GQL fragment that
       should be used when loading this type from GQL. In most cases
        this is just the ID and name, but in some cases we need more
        information. For example, when loading a Run, we also load the Project
        (and extension, the project's Entity).
    2. The data is currently stored in `User.gql` - which gets serialized as an
       untyped string, but can be accessed like a dictionary. See
       `UntypedOpaqueDict` for details.

TODO: Decide if the nested fragments are really necessary. If not, it will
require changing a few of the ops to fetch such info as well
      as updating the mocked tests.

TODO: Refactor how the opaque dict works - I think we can remove it and make
each type feel/look like a a json blob.
"""


@weave_type("UntypedOpaqueDict", True)
class UntypedOpaqueDict:
    """
    UntypedOpaqueDict is a Weave Type that is used to store arbitrary JSON data.
    Unlike `Dict` or `TypedDict`, this Type does not need to define the keys/fields.
    This is useful in particular for storing GQL responses where the response schema
    may change over time. Usage:

    # From JSON String
    d = UntypedOpaqueDict(json_str='{"a": 1, "b": 2}')
    d["a"]  # 1

    # From Dictionary
    d = UntypedOpaqueDict.from_dict({"a": 1, "b": 2})
    d["a"]  # 1

    Importantly, this will serialize the data as a JSON string, so it can be stored and
    loaded using the Weave Type system.
    """

    json_str: str = field(default="{}")

    @classmethod
    def from_json_dict(cls, json_dict: dict):
        inst = cls(json_str=json.dumps(json_dict, separators=(",", ":")))
        inst._json_dict = json_dict
        return inst

    def get(self, key, default=None):
        return self.json_dict.get(key, default)

    def __eq__(self, other):
        return self.json_dict == other.json_dict

    def __getitem__(self, key):
        return self.json_dict[key]

    def __setitem__(self, key, value):
        raise NotImplementedError("UntypedOpaqueDict is immutable")

    def __delitem__(self, key):
        raise NotImplementedError("UntypedOpaqueDict is immutable")

    def __iter__(self):
        return iter(self.json_dict)

    def __len__(self):
        return len(self.json_dict)

    @property
    def json_dict(self):
        if not hasattr(self, "_json_dict"):
            self._json_dict = json.loads(self.json_str)
        return self._json_dict


@dataclass
class GQLTypeMixin:
    gql: UntypedOpaqueDict = field(default_factory=UntypedOpaqueDict)

    @classmethod
    def from_gql(cls, gql_dict: dict):
        return cls(UntypedOpaqueDict.from_json_dict(gql_dict))


@weave_type("org", True)
class Org(GQLTypeMixin):
    REQUIRED_FRAGMENT = f"""
        id
        name
    """


OrgType = Org.WeaveType()  # type: ignore
OrgType = typing.cast(types.Type, OrgType)


@weave_type("entity", True)
class Entity(GQLTypeMixin):
    REQUIRED_FRAGMENT = f"""
        id
        name
    """


EntityType = Entity.WeaveType()  # type: ignore
EntityType = typing.cast(types.Type, EntityType)


@weave_type("user", True)
class User(GQLTypeMixin):
    REQUIRED_FRAGMENT = f"""
        id
        name
    """


UserType = User.WeaveType()  # type: ignore
UserType = typing.cast(types.Type, UserType)


@weave_type("project", True)
class Project(GQLTypeMixin):
    REQUIRED_FRAGMENT = f"""
        id
        name
    """


ProjectType = Project.WeaveType()  # type: ignore
ProjectType = typing.cast(types.Type, ProjectType)


@weave_type("run", True)
class Run(GQLTypeMixin):
    REQUIRED_FRAGMENT = f"""
        id
        name
    """


RunType = Run.WeaveType()  # type: ignore
RunType = typing.cast(types.Type, RunType)


@weave_type("artifactType", True)
class ArtifactType(GQLTypeMixin):
    REQUIRED_FRAGMENT = f"""
        id
        name
    """


ArtifactTypeType = ArtifactType.WeaveType()  # type: ignore
ArtifactTypeType = typing.cast(types.Type, ArtifactTypeType)


@weave_type("artifact", True)  # Name and Class mismatch intention due to weave0
class ArtifactCollection(GQLTypeMixin):
    REQUIRED_FRAGMENT = f"""
        id
        name
    """


ArtifactCollectionType = ArtifactCollection.WeaveType()  # type: ignore
ArtifactCollectionType = typing.cast(types.Type, ArtifactCollectionType)


@weave_type("artifactVersion", True)
class ArtifactVersion(GQLTypeMixin):
    REQUIRED_FRAGMENT = f"""
        id
    """


ArtifactVersionType = ArtifactVersion.WeaveType()  # type: ignore
ArtifactVersionType = typing.cast(types.Type, ArtifactVersionType)


@weave_type(
    "artifactMembership", True
)  # Name and Class mismatch intention due to weave0
class ArtifactCollectionMembership(GQLTypeMixin):
    REQUIRED_FRAGMENT = f"""
        id
    """


ArtifactCollectionMembershipType = ArtifactCollectionMembership.WeaveType()  # type: ignore
ArtifactCollectionMembershipType = typing.cast(
    types.Type, ArtifactCollectionMembershipType
)


@weave_type("artifactAlias", True)
class ArtifactAlias(GQLTypeMixin):
    REQUIRED_FRAGMENT = f"""
        id
    """


ArtifactAliasType = ArtifactAlias.WeaveType()  # type: ignore
ArtifactAliasType = typing.cast(types.Type, ArtifactAliasType)


@weave_type("report", True)
class Report(GQLTypeMixin):
    REQUIRED_FRAGMENT = f"""
        id
    """


ReportType = Report.WeaveType()  # type: ignore
ReportType = typing.cast(types.Type, ReportType)


@weave_type("runQueue", True)
class RunQueue(GQLTypeMixin):
    REQUIRED_FRAGMENT = f"""
        id
    """


RunQueueType = RunQueue.WeaveType()  # type: ignore
RunQueueType = typing.cast(types.Type, RunQueueType)


# Simple types (maybe should be put into primitives?)


@weave_type("link", True)
class Link:
    name: str
    url: str


LinkType = Link.WeaveType()  # type: ignore
LinkType = typing.cast(types.Type, LinkType)



from ..api import op
from .. import weave_types as types
from . import wb_domain_types as wdt
from ..compile_domain import wb_gql_op_plugin
from .wandb_domain_gql import (
    gql_prop_op,
    gql_direct_edge_op,
)

import urllib


# Section 1/6: Tag Getters
# None

# Section 2/6: Root Ops
# None

# Section 3/6: Attribute Getters
gql_prop_op(
    "artifactMembership-versionIndex",
    wdt.ArtifactCollectionMembershipType,
    "versionIndex",
    types.Int(),
)


gql_prop_op(
    "artifactMembership-createdAt",
    wdt.ArtifactCollectionMembershipType,
    "createdAt",
    types.Timestamp(),
)

gql_prop_op(
    "artifactMembership-commitHash",
    wdt.ArtifactCollectionMembershipType,
    "commitHash",
    types.String(),
)

# Section 4/6: Direct Relationship Ops
gql_direct_edge_op(
    "artifactMembership-collection",
    wdt.ArtifactCollectionMembershipType,
    "artifactCollection",
    wdt.ArtifactCollectionType,
)

gql_direct_edge_op(
    "artifactMembership-artifactVersion",
    wdt.ArtifactCollectionMembershipType,
    "artifact",
    wdt.ArtifactVersionType,
)

gql_direct_edge_op(
    "artifactMembership-aliases",
    wdt.ArtifactCollectionMembershipType,
    "aliases",
    wdt.ArtifactAliasType,
    is_many=True,
)

# Section 5/6: Connection Ops
# None

# Section 6/6: Non Standard Business Logic Ops
# None


@op(
    name="artifactMembership-link",
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: """
            versionIndex
            artifactCollection {
                id 
                name 
                defaultArtifactType {
                    id 
                    name 
                    project {
                        id 
                        name 
                        entity {
                            id 
                            name
                        }
                    }
                }
            }
        """,
    ),
)
def artifact_membership_link(
    artifactMembership: wdt.ArtifactCollectionMembership,
) -> wdt.Link:
    return wdt.Link(
        name=f"{artifactMembership.gql['artifactCollection']['name']}:v{artifactMembership.gql['versionIndex']}",
        url=f"/{artifactMembership.gql['artifactCollection']['defaultArtifactType']['project']['entity']['name']}/"
        f"{artifactMembership.gql['artifactCollection']['defaultArtifactType']['project']['name']}/"
        f"artifacts/{urllib.parse.quote(artifactMembership.gql['artifactCollection']['defaultArtifactType']['name'])}/"
        f"{urllib.parse.quote(artifactMembership.gql['artifactCollection']['name'])}"
        f"/v{artifactMembership.gql['versionIndex']}",
    )



import dataclasses
import typing
import wandb
from wandb.sdk.artifacts.artifact_saver import ArtifactSaver
from wandb.sdk.artifacts.public_artifact import Artifact
from wandb.sdk.internal.sender import _manifest_json_from_proto

# from wandb.sdk.internal.artifact_saver import ArtifactSaver # This symbol moved after our pinned version
# from wandb.sdk.internal.sender import _manifest_json_from_proto # This symbol moved after our pinned version
from wandb.sdk.interface.interface import InterfaceBase

from weave import wandb_client_api
from weave.wandb_interface.wandb_lite_run import InMemoryLazyLiteRun


@dataclasses.dataclass
class WeaveWBArtifactURIComponents:
    entity_name: str
    project_name: str
    artifact_name: str
    version_str: str


def write_artifact_to_wandb(
    artifact: wandb.Artifact,
    project_name: str,
    entity_name: typing.Optional[str] = None,
    additional_aliases: list = [],
    *,
    _lite_run: typing.Optional[InMemoryLazyLiteRun] = None,
) -> WeaveWBArtifactURIComponents:
    # Extract Artifact Attributes
    artifact_name = artifact.name
    artifact_type_name = artifact.type

    assert artifact_name is not None
    assert artifact_type_name is not None

    wandb_client_api.assert_wandb_authenticated()

    # Here we get the default entity if none is provided.
    # When we support saving dashboards to target entities,
    # we may want to rework this code path so that the caller
    # must provide an entity name.
    entity_name = entity_name or wandb_client_api.wandb_public_api().default_entity

    if _lite_run is None:
        lite_run = InMemoryLazyLiteRun(
            entity_name, project_name, group="weave_artifact_pushers", _hide_in_wb=True
        )
    else:
        lite_run = _lite_run

    # Ensure the artifact type exists
    lite_run.i_api.create_artifact_type(
        artifact_type_name=artifact_type_name,
        entity_name=lite_run.run.entity,
        project_name=lite_run.run.project,
    )

    # Finalize the artifact and construct the manifest.
    artifact.finalize()
    manifest_dict = _manifest_json_from_proto(
        InterfaceBase()._make_artifact(artifact).manifest  # type: ignore[abstract, attr-defined]
    )

    # Save the Artifact and the associated files
    saver = ArtifactSaver(
        api=lite_run.i_api,
        digest=artifact.digest,
        manifest_json=manifest_dict,
        file_pusher=lite_run.pusher,
        is_user_created=False,
    )
    res = saver.save(
        type=artifact_type_name,
        name=artifact_name,
        client_id=artifact._client_id,
        sequence_client_id=artifact._sequence_client_id,
        metadata=artifact.metadata,
        description=artifact.description,
        aliases=["latest"] + additional_aliases,
        use_after_commit=False,
    )

    lite_run.finish()

    if res is not None:
        art = Artifact.from_id(res["id"], wandb_client_api.wandb_public_api().client)
        if art is not None:
            commit_hash = art.commit_hash

    # Return the URI of the artifact
    return WeaveWBArtifactURIComponents(
        entity_name=lite_run.run.entity,
        project_name=lite_run.run.project,
        artifact_name=artifact_name,
        version_str=commit_hash,
    )



import weave

from . import file_local


def test_file_browsing():
    test_dir = weave.ops.local_path("./testdata/").path("test_dir")
    assert test_dir.type == file_local.LocalDirType()
    assert weave.use(test_dir.path("b.txt").contents()) == "howdy\n"



# Official interface for interacting with the W&B API. All
# Weave interactions with the Weave API should go through this
# module.

import typing
import graphql
import gql
import aiohttp
import contextlib
import contextvars

from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.requests import RequestsHTTPTransport
from requests.auth import HTTPBasicAuth

from . import engine_trace
from . import environment as weave_env
from . import wandb_client_api

# Importing at the top-level namespace so other files can import from here.
from .context_state import WandbApiContext, _wandb_api_context

tracer = engine_trace.tracer()  # type: ignore


def set_wandb_api_context(
    user_id: typing.Optional[str],
    api_key: typing.Optional[str],
    headers: typing.Optional[dict],
    cookies: typing.Optional[dict],
) -> typing.Optional[contextvars.Token[typing.Optional[WandbApiContext]]]:
    cur_ctx = get_wandb_api_context()
    if cur_ctx:
        # WANDB API context is only allowed to be set once per thread, since we
        # need to use thread local storage to communicate the context to the wandb
        # lib right now.
        return None
    wandb_client_api.set_wandb_thread_local_api_settings(api_key, cookies, headers)
    return _wandb_api_context.set(WandbApiContext(user_id, api_key, headers, cookies))


def reset_wandb_api_context(
    token: typing.Optional[contextvars.Token[typing.Optional[WandbApiContext]]],
) -> None:
    if token is None:
        return
    wandb_client_api.reset_wandb_thread_local_api_settings()
    _wandb_api_context.reset(token)


@contextlib.contextmanager
def wandb_api_context(
    ctx: typing.Optional[WandbApiContext],
) -> typing.Generator[None, None, None]:
    if ctx:
        token = set_wandb_api_context(
            ctx.user_id, ctx.api_key, ctx.headers, ctx.cookies
        )
    try:
        yield
    finally:
        if ctx:
            reset_wandb_api_context(token)


def get_wandb_api_context() -> typing.Optional[WandbApiContext]:
    return _wandb_api_context.get()


def init() -> typing.Optional[contextvars.Token[typing.Optional[WandbApiContext]]]:
    cookie = weave_env.weave_wandb_cookie()
    if cookie:
        cookies = {"wandb": cookie}
        headers = {"use-admin-privileges": "true", "x-origin": "https://app.wandb.test"}
        return set_wandb_api_context("admin", None, headers, cookies)
    api_key = weave_env.weave_wandb_api_key()
    if api_key:
        return set_wandb_api_context("admin", api_key, None, None)
    return None


@contextlib.contextmanager
def from_environment() -> typing.Generator[None, None, None]:
    token = init()
    try:
        yield
    finally:
        if token:
            reset_wandb_api_context(token)


class WandbApiAsync:
    def __init__(self) -> None:
        self.connector = aiohttp.TCPConnector(limit=50)

    async def query(
        self, query: graphql.DocumentNode, **kwargs: typing.Any
    ) -> typing.Any:
        wandb_context = get_wandb_api_context()
        headers = None
        cookies = None
        auth = None
        if wandb_context is not None:
            headers = wandb_context.headers
            cookies = wandb_context.cookies
            if wandb_context.api_key is not None:
                auth = aiohttp.BasicAuth("api", wandb_context.api_key)
        url_base = weave_env.wandb_base_url()
        transport = AIOHTTPTransport(
            url=url_base + "/graphql",
            client_session_args={
                "connector": self.connector,
                "connector_owner": False,
            },
            headers=headers,
            cookies=cookies,
            auth=auth,
        )
        # Warning: we do not use the recommended context manager pattern, because we're
        # using connector_owner to tell the session not to close our connection pool.
        # There is a bug in aiohttp that causes session close to hang for the ssl_close_timeout
        # which is 10 seconds by default. See issue: https://github.com/graphql-python/gql/issues/381
        # Closing the session just closes the connector, which we don't want anyway, so we don't
        # bother.
        client = gql.Client(transport=transport, fetch_schema_from_transport=False)
        session = await client.connect_async(reconnecting=False)  # type: ignore
        return await session.execute(query, kwargs)

    SERVER_INFO_QUERY = gql.gql(
        """
        query ServerInfo {
            serverInfo {
            frontendHost
            }
        }
        """
    )

    async def server_info(self) -> typing.Any:
        return await self.query(self.SERVER_INFO_QUERY)

    ARTIFACT_MANIFEST_QUERY = gql.gql(
        """
        query ArtifactManifest(
            $entityName: String!,
            $projectName: String!,
            $name: String!
        ) {
            project(name: $projectName, entityName: $entityName) {
                artifact(name: $name) {
                    currentManifest {
                        id
                        file {
                            directUrl
                        }
                    }
                }
            }
        }
        """
    )

    async def artifact_manifest_url(
        self, entity_name: str, project_name: str, name: str
    ) -> typing.Optional[str]:
        try:
            result = await self.query(
                self.ARTIFACT_MANIFEST_QUERY,
                entityName=entity_name,
                projectName=project_name,
                name=name,
            )
        except gql.transport.exceptions.TransportQueryError as e:
            return None
        project = result["project"]
        if project is None:
            return None
        artifact = project["artifact"]
        if artifact is None:
            return None
        current_manifest = artifact["currentManifest"]
        if current_manifest is None:
            return None
        file = current_manifest["file"]
        if file is None:
            return None
        return file["directUrl"]


class WandbApi:
    def query(self, query: graphql.DocumentNode, **kwargs: typing.Any) -> typing.Any:
        wandb_context = get_wandb_api_context()
        headers = None
        cookies = None
        auth = None
        if wandb_context is not None:
            headers = wandb_context.headers
            cookies = wandb_context.cookies
            if wandb_context.api_key is not None:
                auth = HTTPBasicAuth("api", wandb_context.api_key)
        url_base = weave_env.wandb_base_url()
        transport = RequestsHTTPTransport(
            url=url_base + "/graphql", headers=headers, cookies=cookies, auth=auth
        )
        # Warning: we do not use the recommended context manager pattern, because we're
        # using connector_owner to tell the session not to close our connection pool.
        # There is a bug in aiohttp that causes session close to hang for the ssl_close_timeout
        # which is 10 seconds by default. See issue: https://github.com/graphql-python/gql/issues/381
        # Closing the session just closes the connector, which we don't want anyway, so we don't
        # bother.
        client = gql.Client(transport=transport, fetch_schema_from_transport=False)
        session = client.connect_sync()  # type: ignore
        return session.execute(query, kwargs)

    SERVER_INFO_QUERY = gql.gql(
        """
        query ServerInfo {
            serverInfo {
            frontendHost
            }
        }
        """
    )

    def server_info(self) -> typing.Any:
        return self.query(self.SERVER_INFO_QUERY)

    ARTIFACT_MANIFEST_QUERY = gql.gql(
        """
        query ArtifactManifest(
            $entityName: String!,
            $projectName: String!,
            $name: String!
        ) {
            project(name: $projectName, entityName: $entityName) {
                artifact(name: $name) {
                    currentManifest {
                        id
                        file {
                            directUrl
                        }
                    }
                }
            }
        }
        """
    )

    def artifact_manifest_url(
        self, entity_name: str, project_name: str, name: str
    ) -> typing.Optional[str]:
        try:
            result = self.query(
                self.ARTIFACT_MANIFEST_QUERY,
                entityName=entity_name,
                projectName=project_name,
                name=name,
            )
        except gql.transport.exceptions.TransportQueryError as e:
            return None
        project = result["project"]
        if project is None:
            return None
        artifact = project["artifact"]
        if artifact is None:
            return None
        current_manifest = artifact["currentManifest"]
        if current_manifest is None:
            return None
        file = current_manifest["file"]
        if file is None:
            return None
        return file["directUrl"]


async def get_wandb_api() -> WandbApiAsync:
    return WandbApiAsync()


def get_wandb_api_sync() -> WandbApi:
    return WandbApi()



from . import context_state as _context
from . import logs as _logging

_loading_builtins_token = _context.set_loading_built_ins()

from . import weave_types as types

# this patches the `Type` class to enable the make method - needed due to circular references
from . import make_type as _make_type
from . import ops
from . import codify
from .graph import Node  # used as a type in op definitions
from .show import show, show_url
from .api import *

from . import panels
from .panel import Panel

from .errors import *

# TODO: Don't expose
from .panel_util import make_node

from . import wandb_api as _wandb_api

from .core_types import *

from .panels_py import *

_wandb_api.init()

_context.clear_loading_built_ins(_loading_builtins_token)

# Wow, this works! you can do just "weave" in a notebook and render
# something. Maybe render ecosystem panel?
# from .ecosystem import ecosystem


# def _ipython_display_():
#     return show(ecosystem())

__version__ = "0.24.1"



import dataclasses
import typing

import weave
from .. import panel
from .. import panel_util
from .. import graph


@weave.type()
class ColorConfig:
    pass


@weave.type()
class Color(panel.Panel):
    id = "Color"
    config: typing.Optional[ColorConfig] = dataclasses.field(
        default_factory=lambda: None
    )



import dataclasses
import datetime
import inspect
import os
import re
import sys
import typing
import uuid
import logging


from ..wandb_interface.wandb_stream_table import StreamTable
from .. import errors
from .. import graph

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class ExecutionResult:
    """A dataclass representing the result of a function call.  This is used to pass the result of a function call
    to the monitor callback.
    """

    start_datetime: datetime.datetime
    end_datetime: datetime.datetime
    args: tuple[typing.Any, ...]
    kwargs: dict[str, typing.Any]
    output: typing.Any
    exception: typing.Optional[Exception]
    id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))

    def get(self) -> typing.Any:
        if self.exception is not None:
            raise self.exception
        return self.output


@dataclasses.dataclass
class MonitorRecord:
    _execution_result: ExecutionResult
    _input_preprocessor: typing.Callable[..., typing.Any]
    _output_postprocessor: typing.Callable[..., typing.Any] = lambda x: x
    _additional_data: dict = dataclasses.field(default_factory=dict)
    _on_log: typing.Optional[typing.Callable[..., None]] = None

    @property
    def id(self) -> str:
        return self._execution_result.id

    def get(self) -> typing.Any:
        return self._execution_result.get()

    def add_data(self, data: dict) -> None:
        self._additional_data.update(data)

    def finalize(self) -> None:
        if self._on_log is not None:
            self._on_log(self)
            self._on_log = None
        else:
            logger.warning("MonitorRecord already finalized")

    @property
    def inputs(self) -> dict:
        return self._input_preprocessor(
            *self._execution_result.args, **self._execution_result.kwargs
        )

    @property
    def output(self) -> typing.Any:
        return self._output_postprocessor(self._execution_result.output)

    def as_dict(self) -> dict:
        exception = self._execution_result.exception
        return {
            "result_id": self._execution_result.id,
            "start_datetime": self._execution_result.start_datetime,
            "end_datetime": self._execution_result.end_datetime,
            "latency_ms": (
                self._execution_result.end_datetime.timestamp()
                - self._execution_result.start_datetime.timestamp()
            )
            * 1000,
            "inputs": self.inputs,
            "output": self.output,
            "exception": (
                f"{type(exception).__name__}: {str(exception)}"
                if exception is not None
                else None
            ),
            **self._additional_data,
        }

    def __repr__(self) -> str:
        return f"MonitorRecord({self._execution_result.id})"


def monitor(
    stream_name: typing.Optional[str] = None,
    *,
    project_name: str,
    entity_name: str,
    input_preprocessor: typing.Optional[
        typing.Callable[..., dict[str, typing.Any]]
    ] = None,
    output_postprocessor: typing.Optional[
        typing.Callable[..., dict[str, typing.Any]]
    ] = None,
    auto_log: bool = True,
    raise_on_error: bool = True,
) -> typing.Callable[..., typing.Callable[..., MonitorRecord]]:
    """Monitor a function.  This is a function decorator for performantely monitoring predictions during inference.
    Specify an input_preprocessor or output_postprocessor to have more control over what gets logged.  Each
    callbable should return a dictionary with scalar values or wandb Media objects.

    Arguments:
        stream_name (typing.Optional[str]): The name of the stream to log to. If None, the stream name will be
            inferred from the function name.
        project_name str:  The name of the W&B Project to log to.
        entity_name str:  The name of the W&B Entity to log to.
        input_preprocessor (typing.Callable[..., typing.Any]):  A function that takes the kwargs of the decorated function and
            returns a dictionary of key value pairs to log.
        output_postprocessor (typing.Callable[..., typing.Any]):  A function that takes the return value of the decorated function
            and returns a dictionary or single value to log.
        auto_log (bool):  If True, the function will automatically log the inputs and outputs of the function call. Set this
            to False if you want to add additional data to the log after the function call.
        raise_on_error (bool):  If True, the function will raise an exception if the function call raises an exception.
            If False, the function will return a MonitorRecord with the exception set.
    """

    def decorator(
        fn: typing.Callable[..., typing.Any]
    ) -> typing.Callable[..., MonitorRecord]:
        return MonitoredFunction(
            fn=fn,
            stream_name=stream_name or _get_default_fn_name(fn),
            project_name=project_name,
            entity_name=entity_name,
            input_preprocessor=input_preprocessor,
            output_postprocessor=output_postprocessor,
            auto_log=auto_log,
            raise_on_error=raise_on_error,
        )

    return decorator


class MonitoredFunction:
    _fn: typing.Callable[..., typing.Any]
    _tracked_fn: typing.Callable[..., ExecutionResult]
    _stream_table: StreamTable
    _input_preprocessor: typing.Callable[..., dict[str, typing.Any]]
    _output_preprocessor: typing.Callable[..., dict[str, typing.Any]]
    _auto_log: bool
    _disabled: bool
    _raise_on_error: bool

    def __init__(
        self,
        *,
        fn: typing.Callable[..., typing.Any],
        stream_name: str,
        project_name: str,
        entity_name: str,
        input_preprocessor: typing.Optional[
            typing.Callable[..., dict[str, typing.Any]]
        ] = None,
        output_postprocessor: typing.Optional[
            typing.Callable[..., dict[str, typing.Any]]
        ] = None,
        auto_log: bool = True,
        raise_on_error: bool = True,
    ):
        self._fn = fn
        self._tracked_fn = _track(raise_on_error)(fn)
        self._disabled = False
        self._auto_log = auto_log
        self._raise_on_error = raise_on_error

        try:
            self._stream_table = StreamTable(
                table_name=stream_name,
                project_name=project_name,
                entity_name=entity_name,
            )
        except errors.WeaveWandbAuthenticationException:
            self._disabled = True
            logger.error("Monitoring disabled because WANDB_API_KEY is not set.")
            print("Couldn't find W&B API key, disabling monitoring.", file=sys.stderr)
            print(
                "Set the WANDB_API_KEY env variable to enable monitoring.",
                file=sys.stderr,
            )

        self._input_preprocessor = input_preprocessor or self._default_input_processor
        self._output_postprocessor = output_postprocessor or (lambda x: x)

    def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        execution_record = self._tracked_fn(*args, **kwargs)
        monitor_record = MonitorRecord(
            execution_record,
            self._input_preprocessor,
            self._output_postprocessor,
        )
        if not self._disabled and self._auto_log:
            self._direct_log(monitor_record)
        else:
            monitor_record._on_log = self._direct_log
        return monitor_record

    # to support instance methods
    def __get__(self, instance: typing.Any, owner: typing.Any) -> "MonitoredFunction":
        if instance is None:
            return self
        else:
            self._fn = self._fn.__get__(instance, owner)
            self._tracked_fn = _track(self._raise_on_error)(self._fn)
            return self

    def _direct_log(self, record: MonitorRecord) -> None:
        self._stream_table.log(record.as_dict())

    def _default_input_processor(self, *args: typing.Any, **kwargs: typing.Any) -> dict:
        arg_dict = _arguments_to_dict(self._fn, args, kwargs)
        if "self" in arg_dict:
            del arg_dict["self"]
        return arg_dict

    def rows(self) -> graph.Node:
        return self._stream_table.rows()


def _arguments_to_dict(
    fn: typing.Callable, args: tuple[typing.Any, ...], kwargs: dict[str, typing.Any]
) -> dict[str, typing.Any]:
    """Given a function and arguments used to call that function, return a dictionary of key value pairs
    that are the bound arguments to the function.
    """
    sig = inspect.signature(fn)
    bound_params = sig.bind(*args, **kwargs)
    bound_params.apply_defaults()
    return dict(bound_params.arguments)


def _sanitize_name(name: str) -> str:
    return re.sub(r"[^a-zA-Z0-9_]", "_", name)


def _get_default_fn_name(fn: typing.Callable) -> str:
    fn_file_name = _sanitize_name(os.path.basename(inspect.getfile(fn)))
    if fn_file_name.endswith(".py"):
        fn_file_name = fn_file_name[:-3]
    fn_name = _sanitize_name(fn.__name__)

    return f"{fn_file_name}-{fn_name}"


def _track(
    raise_on_error: bool = True,
) -> typing.Callable[..., typing.Callable[..., ExecutionResult]]:
    """A low-level utility for tracking function executions."""

    def function_monitor_decorator(
        fn: typing.Callable[..., typing.Any]
    ) -> typing.Callable[..., ExecutionResult]:
        def wrapped_fn(*args: typing.Any, **kwargs: typing.Any) -> "ExecutionResult":
            output = None
            exception = None
            args = args
            kwargs = kwargs
            try:
                start_datetime = datetime.datetime.now()
                output = fn(*args, **kwargs)
                end_datetime = datetime.datetime.now()
            except Exception as e:
                end_datetime = datetime.datetime.now()
                if raise_on_error:
                    raise e
                exception = e
            return ExecutionResult(
                start_datetime=start_datetime,
                end_datetime=end_datetime,
                args=args,
                kwargs=kwargs,
                output=output,
                exception=exception,
            )

        return wrapped_fn

    return function_monitor_decorator



import logging
from .. import weave_types as types
from ..api import op
from . import wb_domain_types as wdt
from ..wandb_client_api import wandb_gql_query
from ..language_features.tagging import tagged_value_type
from .. import engine_trace
from .. import errors


def _wbgqlquery_output_type(input_types: dict[str, types.Type]) -> types.Type:
    ot = input_types["alias_list"]
    if isinstance(ot, types.Const) and isinstance(ot.val, list):
        return types.TypedDict({alias: types.TypedDict() for alias in ot.val})
    return types.Dict(types.String(), types.TypedDict({}))


# This op replaces all domain root ops in the graph during the compilation step.
# It executes a GQL query (that is constructed inside of `compile_domain.py`)
# and returns the results as a weave type.
@op(
    name="gqlroot-wbgqlquery",
    input_type={"query_str": types.String(), "alias_list": types.List(types.String())},
    output_type=_wbgqlquery_output_type,
    pure=False,
)
def wbgqlquery(query_str, alias_list):
    tracer = engine_trace.tracer()
    with tracer.trace("wbgqlquery:public_api"):
        logging.info("Executing GQL query: %s", query_str)
        gql_payload = wandb_gql_query(query_str)
    for alias in alias_list:
        if alias not in gql_payload:
            raise errors.WeaveGQLExecuteMissingAliasError(
                f"Alias {alias} not found in query results"
            )
    return gql_payload


def _querytoobj_output_type(input_types: dict[str, types.Type]) -> types.Type:
    ot = input_types["output_type"]
    if isinstance(ot, types.Const) and isinstance(ot.val, types.Type):
        return ot.val
    return types.Any()


@op(
    name="gqlroot-querytoobj",
    input_type={
        "result_dict": types.TypedDict({}),
        "result_key": types.String(),
        "output_type": types.TypeType(),
    },
    output_type=_querytoobj_output_type,
)
def querytoobj(result_dict, result_key, output_type):
    if isinstance(output_type, tagged_value_type.TaggedValueType):
        output_type = output_type.value
    if output_type.instance_class is None or not issubclass(
        output_type.instance_class, wdt.GQLTypeMixin
    ):
        raise ValueError(
            f"Invalid output type for gqlroot-querytoobj, must be a GQLTypeMixin, got {output_type}"
        )
    res_gql = result_dict[result_key]
    if res_gql == None:
        return None
    res = output_type.instance_class.from_gql(res_gql)
    return res
